/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2011 Centre Tecnologic de Telecomunicacions de Catalunya (CTTC)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Jaume Nin <jaume.nin@cttc.cat>
 *
 * Modifications by:
 *      Jason Quinlan j.quinlan@cs.ucc.ie
 *      Ahmed Zahran a.zahran@cs.ucc.ie
 *      Andrew Nash (MISL Summer of Code 2019)
 *      Kieran Horgan (MISL Summer of Code 2019)
 */

// sample execution code
// ./waf --run "scratch/lena-simple-epc-tap-nas-v2.2.2 --numberOfNodes=2 --seedValue=1 --mmwaveScheduler=tti --simTime=7.00 --distance1=10.0 --distance2=14.0 --distance3=18.0 --distance4=22.0 --distance5=26.0 --distance6=30.0 --symPerSf=96 --sfPeriod=400 --useMobility=false --pcapGenerate=false --tcp=true --trafficGeneration=true"
 
  
 // or simply - create traffic using TCP and generate pcap files
 
 // ./waf --run "scratch/lena-simple-epc-tap-nas-v2.2.2 --pcapGenerate=true --tcp=true --trafficGeneration=true"

#include "ns3/lte-helper.h"
#include "ns3/epc-helper.h"
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/internet-module.h"
#include "ns3/mobility-module.h"
#include "ns3/lte-module.h"
#include "ns3/applications-module.h"
#include "ns3/point-to-point-helper.h"
#include "ns3/config-store.h"
//#include "ns3/gtk-config-store.h"


#include "ns3/tap-bridge-helper.h"
#include "ns3/csma-helper.h"
#include "ns3/netanim-module.h"

#include <iostream>
#include <fstream>
#include <sstream>

#include <chrono>

using namespace ns3;
//using namespace mmwave;
/**
 * Sample simulation script for LTE+EPC.
 */

// from 'lena-simple-tap.cc'
NS_LOG_COMPONENT_DEFINE ("LenaSimpleTap2UE");



class MyAppTag : public Tag
{
public:
	MyAppTag ()
	{
	}

	MyAppTag (Time sendTs) : m_sendTs (sendTs)
	{
	}

  static TypeId GetTypeId (void)
  {
    static TypeId tid = TypeId ("ns3::MyAppTag")
      .SetParent<Tag> ()
      .AddConstructor<MyAppTag> ();
    return tid;
  }

  virtual TypeId  GetInstanceTypeId (void) const
  {
    return GetTypeId ();
  }

  virtual void  Serialize (TagBuffer i) const
  {
  	i.WriteU64 (m_sendTs.GetNanoSeconds());
  }

  virtual void  Deserialize (TagBuffer i)
  {
  	m_sendTs = NanoSeconds (i.ReadU64 ());
  }

  virtual uint32_t  GetSerializedSize () const
  {
  	return sizeof (m_sendTs);
  }

  virtual void Print (std::ostream &os) const
  {
  	std::cout << m_sendTs;
  }

	Time m_sendTs;
};



class MyApp : public Application
{
public:

  MyApp ();
  virtual ~MyApp();
  void ChangeDataRate (DataRate rate);
  void Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);



private:
  virtual void StartApplication (void);
  virtual void StopApplication (void);

  void ScheduleTx (void);
  void SendPacket (void);

  Ptr<Socket>     m_socket;
  Address         m_peer;
  uint32_t        m_packetSize;
  uint32_t        m_nPackets;
  DataRate        m_dataRate;
  EventId         m_sendEvent;
  bool            m_running;
  uint32_t        m_packetsSent;
};

MyApp::MyApp ()
  : m_socket (0),
    m_peer (),
    m_packetSize (0),
    m_nPackets (0),
    m_dataRate (0),
    m_sendEvent (),
    m_running (false),
    m_packetsSent (0)
{
}

MyApp::~MyApp()
{
  m_socket = 0;
}

void
MyApp::Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
  m_socket = socket;
  m_peer = address;
  m_packetSize = packetSize;
  m_nPackets = nPackets;
  m_dataRate = dataRate;
}

void
MyApp::ChangeDataRate (DataRate rate)
{
	m_dataRate = rate;
}

void
MyApp::StartApplication (void)
{
  m_running = true;
  m_packetsSent = 0;
  m_socket->Bind ();
  m_socket->Connect (m_peer);
  SendPacket ();
}

void
MyApp::StopApplication (void)
{
  m_running = false;

  if (m_sendEvent.IsRunning ())
    {
      Simulator::Cancel (m_sendEvent);
    }

  if (m_socket)
    {
      m_socket->Close ();
    }
}

void
MyApp::SendPacket (void)
{
	Ptr<Packet> packet = Create<Packet> (m_packetSize);
	MyAppTag tag (Simulator::Now ());

	m_socket->Send (packet);
  	if (++m_packetsSent < m_nPackets)
	{
	    ScheduleTx ();
	}
}



void
MyApp::ScheduleTx (void)
{
	if (m_running)
	{
		Time tNext (Seconds (m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate ())));
		m_sendEvent = Simulator::Schedule (tNext, &MyApp::SendPacket, this);
	}
}

static void
CwndChange (Ptr<OutputStreamWrapper> stream, uint32_t oldCwnd, uint32_t newCwnd)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << oldCwnd << "\t" << newCwnd << std::endl;
}

static void
RttChange (Ptr<OutputStreamWrapper> stream, Time oldRtt, Time newRtt)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << oldRtt.GetSeconds () << "\t" << newRtt.GetSeconds () << std::endl;
}


static void Rx (Ptr<OutputStreamWrapper> stream, Ptr<const Packet> packet, const Address &from)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << packet->GetSize()<< std::endl;
}



/*
* from lena-x2-handover-measures.cc
* this prints out the "eNB CellId", "IMSI" and "RNTI" for each connected UE
* the "eNB CellId" is the ID of the cell tower
*
* International Mobile Subscriber Identity (IMSI)
* the "IMSI" is a number incremental from 1 in the order in which the UE are allocated to the system
*
* Radio Network Temporary Identifier (RNTI)
* the "RNTI" is a random number for each UE, typically in reverse order to which the UE are allocated to the system
* i.e. first allocated has the highest RNTI, but not always
*/
void
NotifyConnectionEstablishedEnb (std::string context,
                                uint64_t imsi,
                                uint16_t cellid,
                                uint16_t rnti)
{
//  std::cout << context
  std::cout << "eNB CellId " << cellid
            << " IMSI " << imsi
            << " RNTI " << rnti
            << std::endl;
}
std::chrono::time_point<std::chrono::system_clock> t_start;
// function to stop the simulation, by reading from an input file in which we can change the value from 0 (run) to 1 (stop)
void checkSimulation(double simTime)
{

    // read from the input file
    std::ifstream inputFile("lteStopFile.txt");
    std::string line;

    // if the file exists and is open
    if ( inputFile.is_open() )
    {
        // get the first file from the file
        std::getline(inputFile, line);

        // output a print line for us to review
        auto t_cur = std::chrono::high_resolution_clock::now();
std::cout
              << "Wall clock time passed: "
              << std::chrono::duration<double, std::milli>(t_cur-t_start).count()/1000
              << " s   ";
        std::cout << "At time instance " << simTime << " the content of the file is " << line <<  "\n";

        // if the input from the file is 1
        if ( line == "1" )
        {
            // output a priont line for us to review
            std::cout << "Kill the clients at " << simTime << "\n";
            // stop the simulation
            Simulator::Stop(Simulator::Now());
        }
        // close the file
        inputFile.close();

    }
    // if the file does not exist or can not be opened
    else
    {
        std::cout << "File not opening" <<  "\n";
    }
}

// function to reset the contents of the input file used to stop the simulation
void writeSimulationValue()
{
    // read from the input file
    std::ofstream inputFile("lteStopFile.txt");
    std::string line;

    // if the file exists and is open
    if ( inputFile.is_open() )
    {
        // write a '0' to the input file, by overwriting the existing content
        inputFile << "0";
        // close the file
        inputFile.close();

    }
    // if the file does not exist or can not be opened
    else
    {
        std::cout << "File not opening" <<  "\n";
        return;
    }
}


// function to get the bandwidth and other metrics from the eNB and UE
void checkBandwidth(Ptr<NetDevice> ueDevice, Ptr<NetDevice> enbDevice)
{

    Ptr<LteEnbNetDevice> enbLteDevice = enbDevice->GetObject<LteEnbNetDevice> ();
    Ptr<LteEnbPhy> enbPhy = enbLteDevice->GetPhy ();

    Ptr<LteUeNetDevice> ueLteDevice = ueDevice->GetObject<LteUeNetDevice> ();
    Ptr<LteUeRrc> ueRrc = ueLteDevice->GetRrc ();
    Ptr<LteUePhy> uePhy = ueLteDevice->GetPhy ();

    std::cout << "\n";
    uint16_t ueCellId = ueRrc->GetCellId ();
    std::cout << "ueCellId " << ueCellId << "\n";
    uint16_t enbCellId = enbLteDevice->GetCellId ();
    std::cout << "enbCellId " << enbCellId << "\n";
    uint8_t ueDlBandwidth = ueRrc->GetDlBandwidth ();
    std::cout << "UE-1 download - # RB = " << (uint32_t) ueDlBandwidth << ", => " << (((uint32_t) ueDlBandwidth) * .2) << "MHz\n";
    uint8_t enbDlBandwidth = enbLteDevice->GetDlBandwidth ();
    std::cout << "ENB download " << (uint32_t) enbDlBandwidth << "\n";
    uint8_t ueUlBandwidth = ueRrc->GetUlBandwidth ();
    std::cout << "UE-1 upload " << (uint32_t) ueUlBandwidth << "\n";
    uint8_t enbUlBandwidth = enbLteDevice->GetUlBandwidth ();
    std::cout << "ENB upload " << (uint32_t) enbUlBandwidth << "\n";
    uint8_t ueDlEarfcn = ueRrc->GetDlEarfcn ();
    std::cout << "ueDlEarfcn " << (uint32_t) ueDlEarfcn << "\n";
    uint8_t enbDlEarfcn = enbLteDevice->GetDlEarfcn ();
    std::cout << "enbDlEarfcn " << (uint32_t) enbDlEarfcn << "\n";
    uint8_t ueUlEarfcn = ueRrc->GetUlEarfcn ();
    std::cout << "ueUlEarfcn " << (uint32_t) ueUlEarfcn << "\n";
    uint8_t enbUlEarfcn = enbLteDevice->GetUlEarfcn ();
    std::cout << "enbUlEarfcn " << (uint32_t) enbUlEarfcn << "\n";
    uint64_t ueImsi = ueLteDevice->GetImsi ();
    std::cout << "ueImsi " << ueImsi << "\n";
    double ueTxPower = uePhy->GetTxPower();
    std::cout << "ueTxPower " << ueTxPower << "\n";
    double enbTxPower = enbPhy->GetTxPower();
    std::cout << "enbTxPower " << enbTxPower << "\n";

}

// change the direction of the UE
static void CourseChange(std::string foo, Ptr<const MobilityModel> mobility)
{
    // get the current simulation time
    Time now = Simulator::Now();
    // get the position of the UE
    Vector pos = mobility->GetPosition();
    // print the co-ordinates of the UE
    std::cout <<"Time: "<<now.GetSeconds()<<" "<<foo<<"POS: x="<<pos.x<< ", y="<<pos.y<<", z="<<pos.z<<std::endl;
}

// define the structure of the LENA architecture
int
main (int argc, char *argv[])
{

    // *****************  LETS DEFINE SOME VARIABLES ************************
    // add some new variables from 'lena-simple-tap.cc'
    uint16_t numberOfNodes = 2;
    uint16_t numberOfStreamingClients = 2;
    uint16_t numberOfeNodeBs = 1;
    double simTime = 10.5;
    double distance[12] = {25, 190, 190, 190, 190, 190, 350, 400, 450, 500, 550, 600};
    double interPacketInterval = 1;
    double txPower = 40.3342375549; // different to D-LiTE
    double noisePower = 10.0;
    uint16_t bandwidth = 100; // different to D-LiTE
    bool useFading = false;
    bool mobileFading = false;
    bool vehicleFading = false;
    std::string LTEscheduler = "pf";
    std::string staticFadingModel = "static_fading_trace_EPA_30mph_20min_25rbg.fad"; // new to D-LiTE
    std::string pedestrianFadingModel = "fading_trace_EPA_3kmph_10m_1ms_25RB.fad"; // new to D-LiTE
    std::string vehicularFadingModel = "fading_trace_EVA_30kmph_20m.fad"; // new to D-LiTE
    //std::string vehicularFadingModel = "vehicular_fading_trace_EVA_30kmph_20min_6rbg.fad";
    uint16_t numRB = 25; // new to D-LiTE
    uint16_t seedValue = 1;
    bool backgroundTraffic = false; // new to D-LiTE
    double LteTxPower = 15.0; // new to D-LiTE
    bool useMobility = true;
    std::string device_DataRate = "10Gb/s"; // different to D-LiTE
    double channel_Delay = 0.001; // different to D-LiTE

    //some new variables to enable tcp/udp traffic generation

    // use tcp
    bool tcp = false;
    // generate pcap
    bool pcapGenerate = false;
    // switch on traffic generation
    bool trafficGeneration = false;
    // Name of file for animation output
    std::string animFile = "ns3-lte-animation.xml";

    // Command line arguments
    // add some new variables from 'lena-simple-tap.cc'
    CommandLine cmd;
    cmd.AddValue("numberOfNodes", "Number of UEs", numberOfNodes);
    cmd.AddValue("numberOfStreamingClients", "Number of UEs", numberOfStreamingClients);
    cmd.AddValue("numberOfeNodeBs", "Number of eNodeBs", numberOfeNodeBs);
    cmd.AddValue("simTime", "Total duration of the simulation [s])", simTime);
    cmd.AddValue("distance1", "Distance between eNBs [m]", distance[0]);
    cmd.AddValue("distance2", "Distance between eNBs [m]", distance[1]);
    cmd.AddValue("distance3", "Distance between eNBs [m]", distance[2]);
    cmd.AddValue("distance4", "Distance between eNBs [m]", distance[3]);
    cmd.AddValue("distance5", "Distance between eNBs [m]", distance[4]);
    cmd.AddValue("distance6", "Distance between eNBs [m]", distance[5]);
    cmd.AddValue("distance7", "Distance between eNBs [m]", distance[6]);
    cmd.AddValue("distance8", "Distance between eNBs [m]", distance[7]);
    cmd.AddValue("distance9", "Distance between eNBs [m]", distance[8]);
    cmd.AddValue("distance10", "Distance between eNBs [m]", distance[9]);
    cmd.AddValue("distance11", "Distance between eNBs [m]", distance[10]);
    cmd.AddValue("distance12", "Distance between eNBs [m]", distance[11]);
    cmd.AddValue("interPacketInterval", "Inter packet interval [ms])", interPacketInterval);
    cmd.AddValue("txPower", "", txPower);
    cmd.AddValue("bandwidth", "", bandwidth);
    cmd.AddValue("noisePower", "", noisePower);
    cmd.AddValue("useFading", "", useFading);
    cmd.AddValue("mobileFading", "", mobileFading);
    cmd.AddValue("vehicleFading", "", vehicleFading);
    cmd.AddValue("LTEscheduler", "", LTEscheduler);
    cmd.AddValue("staticFadingModel", "", staticFadingModel);
    cmd.AddValue("pedestrianFadingModel", "", pedestrianFadingModel);
    cmd.AddValue("vehicularFadingModel", "", vehicularFadingModel);
    cmd.AddValue("numRB", "", numRB); // new
    cmd.AddValue("seedValue", "", seedValue);
    cmd.AddValue("backgroundTraffic", "", backgroundTraffic);
    cmd.AddValue("LteTxPower", "", LteTxPower); // new
    cmd.AddValue("useMobility", "", useMobility); // new
    cmd.AddValue("device_DataRate", "", device_DataRate); // new
    cmd.AddValue("channel_Delay", "", channel_Delay); // new


    //new arguments - 2019
    cmd.AddValue ("tcp", "<true> to enable UDP and <false> to enable UDP", tcp);
    cmd.AddValue ("pcapGenerate", "<true> to enable wireshark pcap files", pcapGenerate);
    cmd.AddValue ("trafficGeneration", "<true> to generate traffic", trafficGeneration);


    cmd.Parse(argc, argv);


    int PacketSize = 2500;
  Config::SetDefault ("ns3::TcpSocketBase::MinRto", TimeValue (MilliSeconds (200)));
  Config::SetDefault ("ns3::Ipv4L3Protocol::FragmentExpirationTimeout", TimeValue (Seconds (1)));
  Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (PacketSize));
  Config::SetDefault ("ns3::TcpSocket::DelAckCount", UintegerValue (1));
  //Config::SetDefault ("ns3::TcpSocket::SndBufSize", UintegerValue (131072*200));
  //Config::SetDefault ("ns3::TcpSocket::RcvBufSize", UintegerValue (131072*200));

  Config::SetDefault ("ns3::TcpSocket::SndBufSize", UintegerValue (131072*180*4));
  Config::SetDefault ("ns3::TcpSocket::RcvBufSize", UintegerValue (131072*180*2));


    // *****************  LETS IMPLEMENT REAL-TIME SIMULATION ************************
    // from 'lena-simple-tap.cc'
    GlobalValue::Bind ("SimulatorImplementationType",
                     StringValue ("ns3::RealtimeSimulatorImpl"));
    GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));

    // set a run value, based on passed in parameter
    std::cout << "Seed value of " << seedValue << "\n";
    SeedManager::SetRun(seedValue);

    // *****************  LETS CREATE SOME ADDITIONAL LOG STATEMENTS **************************
    // Enable Logging
    //LogLevel logLevel = (LogLevel)(LOG_PREFIX_FUNC | LOG_PREFIX_TIME | LOG_LEVEL_ALL);
    //LogComponentEnable("Ipv4L3Protocol", logLevel);
    //LogComponentEnable("PointToPointNetDevice", logLevel);
    //LogComponentEnable("TapBridge", logLevel);
    //LogComponentEnable("LteUeNetDevice", logLevel);
    //LogComponentEnable("LteEnbNetDevice", logLevel);
    //LogComponentEnable("LteNetDevice", logLevel);
    //LogComponentEnable("LteUeRrc", logLevel);
    //LogComponentEnable("LteHelper", logLevel);
    //LogComponentEnable("EpcHelper", logLevel);
    //LogComponentEnable("EpcEnbApplication", logLevel);
    //LogComponentEnable("EpcSgwPgwApplication", logLevel);
    //LogComponentEnable("EpcMme", logLevel);
    //LogComponentEnable("LteEnbRrc", logLevel);
    //LogComponentEnable("Ipv4GlobalRouting", logLevel);
    //LogComponentEnable("PointToPointEpcHelper", logLevel);
    //LogComponentEnable("Ipv4ListRouting", logLevel);
    //LogComponentEnable("RealtimeSimulatorImpl", logLevel);

    std::cout << "**** SETUP ****\n";
    std::cout << "LTE and EPC Creation\n";
    Ptr<LteHelper> lteHelper = CreateObject<LteHelper> ();
    Ptr<PointToPointEpcHelper>  epcHelper = CreateObject<PointToPointEpcHelper> ();
    lteHelper->SetEpcHelper (epcHelper);


    // #################  SCHEDULER ###################

    // from 'lena-simple-tap.cc'
    std::cout << "ENB Scheduler\n";
    //lteHelper->SetSchedulerType("ns3::FdBetFfMacScheduler");

    // Pf by default
    //lteHelper->SetSchedulerType ("ns3::PfFfMacScheduler");
    //lteHelper->SetSchedulerAttribute ("CqiTimerThreshold", UintegerValue (3));

    if (LTEscheduler == "fdbet" ){

        std::cout << "FdBetFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::FdBetFfMacScheduler");

    }
    else if (LTEscheduler == "fdmt" ){

        std::cout << "FdMtFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::FdMtFfMacScheduler");

    }
    else if (LTEscheduler == "fdtbfq" ){

        std::cout << "FdTbfqFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::FdTbfqFfMacScheduler");

    }
    else if (LTEscheduler == "pf" ){

        std::cout << "PfFfMacScheduler scheduler is now active - default setting\n";
        lteHelper->SetSchedulerType("ns3::PfFfMacScheduler");

    }
    else if (LTEscheduler == "pss" ){

        std::cout << "PssFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::PssFfMacScheduler");

    }
    else if (LTEscheduler == "rr" ){

        std::cout << "RrFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::RrFfMacScheduler");

    }
    else if (LTEscheduler == "tdbet" ){

        std::cout << "TdBetFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::TdBetFfMacScheduler");

    }
    else if (LTEscheduler == "tdmt" ){

        std::cout << "TdMtFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::TdMtFfMacScheduler");

    }
    else if (LTEscheduler == "tdtbfq" ){

        std::cout << "TdTbfqFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::TdTbfqFfMacScheduler");

    }
    else if (LTEscheduler == "tta" ){

        std::cout << "TtaFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::TtaFfMacScheduler");

    }
    else if (LTEscheduler == "cqa" ){

        std::cout << "CqaFfMacScheduler scheduler is now active\n";
        lteHelper->SetSchedulerType("ns3::CqaFfMacScheduler");

    }

    // ################# RESOURCE BLOCK ALLOCATION ###################
    std::cout << "Bandwidth value of " << bandwidth << "\n";
    lteHelper->SetEnbDeviceAttribute ("DlBandwidth", UintegerValue (bandwidth));
    lteHelper->SetEnbDeviceAttribute ("UlBandwidth", UintegerValue (bandwidth));



    // ################# PATHLOSS MODEL ###################

    //Pathloss Model
    // FriisPropagationLossModel by default
    // LogDistancePropagationLossModel gives us smaller distances from eNB to edge
    lteHelper->SetAttribute ("PathlossModel", StringValue ("ns3::LogDistancePropagationLossModel"));


    // #################  FADING MODEL ###################

    // if we having fading
    if (useFading){

        std::cout << "Using Fading Model\n";

        lteHelper->SetAttribute ("FadingModel", StringValue ("ns3::TraceFadingLossModel"));
        std::ifstream ifTraceFile;

        // select the correct fading model for mobile
        if (mobileFading){

           // select the vehicular fading model
           if (vehicleFading){

              ifTraceFile.open (vehicularFadingModel.c_str(), std::ifstream::in);
              if (ifTraceFile.good ())
                {

                  std::cout << "Mobile Vehicular Fading Model Loaded\n";
                  lteHelper->SetFadingModelAttribute ("TraceFilename", StringValue (vehicularFadingModel));

                  lteHelper->SetFadingModelAttribute ("TraceLength", TimeValue (Seconds (1200.0)));
                  lteHelper->SetFadingModelAttribute ("SamplesNum", UintegerValue (1200000));
                  lteHelper->SetFadingModelAttribute ("RbNum", UintegerValue (numRB));
                }
                else{
                    std::cout << "Mobile Vehicular Fading Model Failed to Load\n";
                }
            }
            // select the pedestrian fading model
            else {

                ifTraceFile.open (pedestrianFadingModel.c_str(), std::ifstream::in);
                if (ifTraceFile.good ())
                {

                  std::cout << "Mobile Pedestrian Fading Model Loaded\n";

                  // script launched by test.py
                  lteHelper->SetFadingModelAttribute ("TraceFilename", StringValue (pedestrianFadingModel));
                  lteHelper->SetFadingModelAttribute ("TraceLength", TimeValue (Seconds (600.0)));
                  lteHelper->SetFadingModelAttribute ("SamplesNum", UintegerValue (600000));
                  lteHelper->SetFadingModelAttribute ("RbNum", UintegerValue (numRB));
                }
                else{
                    std::cout << "Mobile Pedestrian Fading Model Failed to Load\n";
                }
            }
        }

        // use the slightly moving static pedestrian fading model
        else {

            ifTraceFile.open (staticFadingModel.c_str(), std::ifstream::in);
            if (ifTraceFile.good ())
            {
                 std::cout << "Static Fading Model Loaded\n";
                 lteHelper->SetFadingModelAttribute ("TraceFilename", StringValue (staticFadingModel));
                 lteHelper->SetFadingModelAttribute ("TraceLength", TimeValue (Seconds (1200.0)));
                 lteHelper->SetFadingModelAttribute ("SamplesNum", UintegerValue (1200000));
                 lteHelper->SetFadingModelAttribute ("RbNum", UintegerValue (numRB));
            }
            else{
                std::cout << "Static Fading Model Failed to Load\n";
            }
        }

    }

    // add some background trafic - NOT COMPLETE AS PER VERSION 2.0
    if (backgroundTraffic)
    {
        numberOfStreamingClients=numberOfNodes+1;
    }
    else {
      numberOfStreamingClients=numberOfNodes;
    }


    // #################  CONFIG SETUP ###################

    // existing
    std::cout << "Update config settings\n";
    ConfigStore inputConfig;
    inputConfig.ConfigureDefaults();
    // parse again so you can override default values from the command line
    cmd.Parse(argc, argv);
    Ptr<Node> pgw = epcHelper->GetPgwNode ();
    Names::Add ("PGW", pgw);
    //Config::SetDefault ("ns3::LteUePhy::TxPower", DoubleValue(15.0)); - changed to variable value
    Config::SetDefault ("ns3::LteUePhy::TxPower", DoubleValue(LteTxPower));

    // *****************  CREATE PROTOCOL STACKS **********************
    Ipv4StaticRoutingHelper ipv4RoutingHelper;
    Ipv4ListRoutingHelper list1;
    list1.Add (ipv4RoutingHelper, 10);

    // *****************  LETS CREATE THE INTERNET SIDE OF THE EPC **********************
    // create our containers - code found online
    std::cout << "\n";
    std::cout << "**** SERVER NODE ****\n";
    std::cout << "Create Server nodes\n";
    NodeContainer remoteNodes;
    remoteNodes.Create (2);
    Ptr<Node> leftTapLeg = remoteNodes.Get (0);
    Ptr<Node> remoteHost = remoteNodes.Get (1);
    Names::Add ("Remote.1", remoteNodes.Get (0));
    Names::Add ("Remote.2", remoteNodes.Get (1));

    // Create the Internet
    std::cout << "Create CSMA & Internet Helpers\n";
    CsmaHelper csmaRemote;
    csmaRemote.SetChannelAttribute ("DataRate", DataRateValue (DataRate ("100Gb/s")));

    NetDeviceContainer internetDevicesRemote = csmaRemote.Install (remoteNodes);
    // lets add some more names
    Names::Add ("Remote_Left_CSMA", internetDevicesRemote.Get (0));
    Names::Add ("Remote_Right_CSMA", internetDevicesRemote.Get (1));
    InternetStackHelper internetRemote;

    internetRemote.SetRoutingHelper (list1);
    internetRemote.Install (remoteNodes);
    internetRemote.Reset(); //clear the stack

    Ipv4AddressHelper ipv4Remote;
    ipv4Remote.SetBase ("11.0.0.0", "255.255.255.0");
    // burn the IP address, we use
    ipv4Remote.NewAddress();
    //ipv4Remote.NewAddress();
    Ipv4InterfaceContainer interfacesLeft = ipv4Remote.Assign (internetDevicesRemote);

    // get the translation values between the ueNodes and TAP
    Ipv4Address addr8 = interfacesLeft.GetAddress (0);
    std::cout << "Remote Left IP Address = " << addr8 << ".\n";
    Ipv4Address addr7 = interfacesLeft.GetAddress (1);
    std::cout << "Remote Right IP Address = " << addr7 << ".\n";
		TapBridgeHelper tapBridgeRemote(interfacesLeft.GetAddress (1));

    // (interfacesLeft.GetAddress (1));
    // UseBridge is used for extending a logical bridge into ns3
    tapBridgeRemote.SetAttribute ("Mode", StringValue ("UseBridge"));
    tapBridgeRemote.SetAttribute ("DeviceName", StringValue ("tapNas"));
    if (!trafficGeneration)
      {
        tapBridgeRemote.Install (leftTapLeg, internetDevicesRemote.Get (0));
      }

    std::cout << "\n";
    std::cout << "Create P2P Helper\n";
    PointToPointHelper p2ph;
    p2ph.SetDeviceAttribute ("DataRate", DataRateValue (DataRate (device_DataRate.c_str())));
    p2ph.SetDeviceAttribute ("Mtu", UintegerValue (2500));
    p2ph.SetChannelAttribute ("Delay", TimeValue (Seconds (channel_Delay)));
    Ipv4AddressHelper ipv4h;
    ipv4h.SetBase ("1.0.0.0", "255.255.255.0");
    NetDeviceContainer internetDevices = p2ph.Install (pgw, remoteHost);

    Ipv4InterfaceContainer internetIpIfaces = ipv4h.Assign (internetDevices);
    //new 2019
    ipv4h.NewNetwork ();


    // lets add some more names
    Names::Add ("Remote_P2P", internetDevices.Get (1));
    Names::Add ("Pgw_P2P", internetDevices.Get (0));

    Ipv4Address addr9 = internetIpIfaces.GetAddress (1);
    std::cout << "remote Host IP = " << addr9 << ".\n";
    Ipv4Address addr10 = internetIpIfaces.GetAddress (0);
    std::cout << "pgw IP = " << addr10 << ".\n";

    // *****************  LETS CREATE THE UE AND ENB NODES **********************
    std::cout << "\n";
    std::cout << "Create UE and ENB nodes\n";
    NodeContainer ueNodes;
    NodeContainer enbNodes;
    enbNodes.Create(numberOfeNodeBs);
    Names::Add ("ENB", enbNodes.Get (0));
    ueNodes.Create(numberOfStreamingClients);

    // add some names to the ueNodes
    for (uint16_t i = 0; i < numberOfStreamingClients; i++)
    {
        std::stringstream ueNodesName;
        ueNodesName << "UE" << (i+1) << ".1";
        Names::Add (ueNodesName.str(), ueNodes.Get (i));
    }

    std::cout << "Install Mobility Model\n";
    MobilityHelper mobility;
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.Install (enbNodes);

    // all new to D-LiTE (mobility)
    int boxBound = 280;
    // implement mobilty only if we use a mobile fading model and we want to use mobility
    if (mobileFading && useMobility)
    {

        if (vehicleFading){
            std::cout << "Vehicular mobility ..... "  << ".\n";
            mobility.SetMobilityModel ("ns3::GaussMarkovMobilityModel",
           "Bounds", BoxValue (Box (-1*boxBound, boxBound, -1*boxBound, boxBound, 0, 100)),
          "TimeStep", TimeValue (Seconds (0.5)),
          "Alpha", DoubleValue (0.85),
          "MeanVelocity", StringValue ("ns3::ConstantRandomVariable[Constant=8.33]"),
          "MeanDirection", StringValue ("ns3::UniformRandomVariable[Min=0|Max=6.283185307]"),
          "MeanPitch", StringValue ("ns3::UniformRandomVariable[Min=0.005|Max=0.005]"),
          "NormalVelocity", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=1.0|Bound=12.0]"),
          "NormalDirection", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.2|Bound=3.14]"),
          "NormalPitch", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.0002|Bound=0.0004]"));

        } else {
            std::cout << "Pedestrian mobility ..... "  << ".\n";
            mobility.SetMobilityModel ("ns3::GaussMarkovMobilityModel",
          "Bounds", BoxValue (Box (-1*boxBound, boxBound, -1*boxBound, boxBound, 0, 100)),
          "TimeStep", TimeValue (Seconds (0.5)),
          "Alpha", DoubleValue (0.85),
          "MeanVelocity", StringValue ("ns3::ConstantRandomVariable[Constant=1]"),
          "MeanDirection", StringValue ("ns3::UniformRandomVariable[Min=0|Max=6.283185307]"),
          "MeanPitch", StringValue ("ns3::UniformRandomVariable[Min=0.005|Max=0.005]"),
          "NormalVelocity", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=1.0|Bound=12.0]"),
          "NormalDirection", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.2|Bound=3.14]"),
          "NormalPitch", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.0002|Bound=0.0004]"));

        }

    } else {

        mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel"); // probably no need for this
        Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
        for (uint16_t i = 0; i < numberOfNodes; i++)
            {
                positionAlloc->Add (Vector(distance[i], 0, 0)); // different to D-LiTE
                std::cout << "Mobility allocated for UE-" << i+1 << " at a distance of " << distance[i] << ".\n";
            }

        mobility.SetPositionAllocator(positionAlloc);
    }
    mobility.Install(ueNodes);

    // update mobilty only if we use a mobile fading model and we want to use mobility
    if (mobileFading && useMobility)
    {
        for (uint32_t i = 0; i < ueNodes.GetN (); i++)
        {
            Ptr<MobilityModel> ueMmPtr = (ueNodes.Get(i))->GetObject<MobilityModel>();
            double xPos = (40 + i * 70) * cos(0.785);
            double yPos = (40 + i * 70) * sin(0.785);
            ueMmPtr->SetPosition(Vector(xPos, yPos, 0));
        }
    }


    // existing - Install LTE Devices to the nodes
    std::cout << "Create UE and ENB Containers\n";
    NetDeviceContainer enbLteDevs = lteHelper->InstallEnbDevice (enbNodes);
    // lets add some more names
    Names::Add ("ENB1", enbLteDevs.Get (0));
    NetDeviceContainer ueLteDevs = lteHelper->InstallUeDevice (ueNodes);

    // add some names to the ueLteDevs
    for (uint16_t i = 0; i < numberOfStreamingClients; i++)
    {
        // string conversion using IOStreams
        std::stringstream ueLteDevsName;
        ueLteDevsName << "LTE_UE" << (i+1);
        // lets add some more names
        Names::Add (ueLteDevsName.str(), ueLteDevs.Get (i));
    }

    // Install the IP stack on the UEs
    std::cout << "Install Internet Stack on UEs\n";
    InternetStackHelper internet;
    //internet.Install (ueNodes);

    //install protocol stack to the UEs
    for (uint32_t i = 0; i < ueNodes.GetN (); i++)
    {
       internet.SetRoutingHelper (list1);
       internet.Install (ueNodes.Get(i));
     }
    internet.Reset();

    Ipv4InterfaceContainer ueIpIface;
    ueIpIface = epcHelper->AssignUeIpv4Address (NetDeviceContainer (ueLteDevs));

    // Assign IP address to UEs, and install applications
    for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
    {
        Ptr<Node> ueNode = ueNodes.Get (u);
        // Set the default gateway for the UE
        Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ueNode->GetObject<Ipv4> ());
        ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);
    }

    for (uint16_t i = 0; i < numberOfStreamingClients; i++)
    {
        lteHelper->Attach (ueLteDevs.Get(i), enbLteDevs.Get(0));
    }

    // *****************  LETS CREATE SOME UE'S ************************
    // Let's create the first node
    std::cout << "\n";
    std::cout << "Create IP Address Helper\n";
    ipv4h.SetBase ("18.2.0.0", "255.255.0.0");
    std::cout << "View Network Metrics\n";
    Ipv4Address gateway = epcHelper->GetUeDefaultGatewayAddress ();
    std::cout << "Left UE gateway = " << gateway << ".\n";

    // create the UE nodes
    NodeContainer UE;
    UE.Create (numberOfStreamingClients);

    // creat the csma
    CsmaHelper csma;
    csma.SetChannelAttribute ("DataRate", DataRateValue (DataRate ("100Gb/s")));

    // create the tap bridge
    TapBridgeHelper tapBridge;
    tapBridge.SetAttribute ("Mode", StringValue ("UseBridge"));

    Ipv4InterfaceContainer interfacesUE;


    // add some UEs - based on  the number of devices - this will be one less when background traffic is true
    for (uint16_t i = 0; i < numberOfNodes; i++)
    {
        // string conversion using IOStreams
        std::stringstream ueFirstName;
        std::stringstream ueLeftName;
        std::stringstream ueRightName;
        std::stringstream ueTapName;

        // create the strings
        ueFirstName << "UE" << (i+1) << ".2";
        ueLeftName << "UE" << (i+1) << "_Left_CSMA";
        ueRightName << "UE" << (i+1) << "_Right_CSMA";
        ueTapName << "tap-UE" << (i+1);

        // create this UE
        std::cout << "**** UE-"<<(i+1)<<" ****\n";
        NetDeviceContainer internetDevice = csma.Install (NodeContainer (ueNodes.Get (i), UE.Get (i)));
        internet.Install (UE.Get (i));
        // lets add some more names
        Names::Add (ueFirstName.str(), UE.Get (i));
        Names::Add (ueLeftName.str(), internetDevice.Get (0));
        Names::Add (ueRightName.str(), internetDevice.Get (1));
        interfacesUE = ipv4h.Assign (internetDevice);
        std::cout << "Create Tap Bridge\n";
        
        if (!trafficGeneration)
          {
				    tapBridge.SetAttribute ("DeviceName", StringValue (ueTapName.str()));
            tapBridge.Install (UE.Get (i), internetDevice.Get (1));
          }

    }

    // this will be one less when background traffic is true
    for (uint32_t u = 0; u < numberOfNodes; ++u)
    {
        std::cout << "\n";
        // get the translation values between the ueNodes and TAP
        Ptr<Node> ueNode = ueNodes.Get (u);
        Ptr<Ipv4> ipv4 = ueNode->GetObject<Ipv4> ();
        Ipv4Address addr1 = ipv4->GetAddress (1, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " LTE IP Address = " << addr1 << ".\n";
        Ipv4Address addr2 = ipv4->GetAddress (2, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " LTE-CSMA IP Address = " << addr2 << ".\n";
        Ipv4Address TapAddr = UE.Get (u)->GetObject<Ipv4> ()->GetAddress (1, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " CSMA-TAP IP Address = " << TapAddr << ".\n";

        Ptr<LteUeNetDevice> ueLteDevice = ueLteDevs.Get (u)->GetObject<LteUeNetDevice> ();
        uint64_t ueImsi = ueLteDevice->GetImsi ();
        std::cout << "UE-" << u+1 << " IMSI = " << ueImsi << ".\n";
    }

    // add some background trafic
    if (backgroundTraffic)
    {
        // increase the number of UE
        // numberOfNodes=numberOfNodes+1;
    }

    // *****************  LETS CREATE SOME ROUTING ***********************
    std::cout << "\n";
    std::cout << "Create Routing\n";
    // define some static routes for the REMOTE NODE
    Ptr<Ipv4StaticRouting> remoteHostStaticRouting = ipv4RoutingHelper.GetStaticRouting (remoteHost->GetObject<Ipv4> ());
    // 9.0.0.x (NAS) and 12.0.0.x (Clients) are subnets to LTE and the Clients
    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("11.0.0.0"), Ipv4Mask ("255.255.255.0"), 1);
    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("9.0.0.0"), Ipv4Mask ("255.255.255.0"), Ipv4Address ("11.0.0.1"), 1);

    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("18.2.0.0"), Ipv4Mask ("255.0.0.0"), 2);


    // define some static routes for the PGW
    Ptr<Ipv4StaticRouting> remoteHostStaticRoutingPgw = ipv4RoutingHelper.GetStaticRouting (pgw->GetObject<Ipv4> ());
    remoteHostStaticRoutingPgw->AddNetworkRouteTo (Ipv4Address ("9.0.0.0"), Ipv4Mask ("255.255.255.0"), 2);
    remoteHostStaticRoutingPgw->AddNetworkRouteTo (Ipv4Address ("11.0.0.0"), Ipv4Mask ("255.255.255.0"), 2);


   // only add routing for internal nodes when we want internal traffic
   if (trafficGeneration)
   {
		  // this will be one less when background traffic is true
     for (uint32_t u = 0; u < numberOfNodes; ++u)
     {
          // Set the default gateway for the UE
	      Ptr<Node> ueNode = ueNodes.Get (u);
	      Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ueNode->GetObject<Ipv4> ());
	      ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);
        ueStaticRouting->AddNetworkRouteTo (Ipv4Address ("18.2.0.0"), Ipv4Mask ("255.0.0.0"), 3);

          // only used when we don't use TapBridge
        Ptr<Node> ueNodeRemote = UE.Get (u);
        Ptr<Ipv4StaticRouting> ueStaticRoutingRemote = ipv4RoutingHelper.GetStaticRouting (ueNodeRemote->GetObject<Ipv4> ());
        // ueStaticRoutingRemote->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);

          std::ostringstream subnet;
          subnet<<"18.2.0."<<((u*2)+1);
          std::cout << "IP of gateway " << subnet.str ().c_str () << " for client " << u+1 << "\n";
        ueStaticRoutingRemote->AddNetworkRouteTo (Ipv4Address ("1.0.0.0"), Ipv4Mask ("255.255.255.0"), Ipv4Address (subnet.str ().c_str ()), 1);

       }
    }


    // taken from D-LiTE
    // *****************  DEFINE EPS BEARERS FOR SOME SCHEDULERS ***********************
    // this code was taken from 'lte-test-fdtbfq-ff-mac-scheduler.cc' in lte ns3 test folder
    if (LTEscheduler == "fdtbfq" || (LTEscheduler == "tdtbfq" ) || (LTEscheduler == "pss" )){

        std::cout << "Creating EPS Bearer for " << LTEscheduler << "\n";

        // Activate an EPS bearer on all UEs
        for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
        {
            std::cout << "creating bearer for client " << (u+1) << "\n";
            Ptr<NetDevice> ueDevice = ueLteDevs.Get (u);
            GbrQosInformation qos;
            qos.gbrDl = 750000;  // 750Kbps
            qos.gbrUl = 0;
            qos.mbrDl = 1750000; // 1.75Mbps
            qos.mbrUl = 0;

            enum EpsBearer::Qci q = EpsBearer::GBR_CONV_VIDEO;
            EpsBearer bearer (q, qos);
            lteHelper->ActivateDedicatedEpsBearer (ueDevice, bearer, EpcTft::Default ());
        }
    }


    // *****************  PRODUCE SOME TRACES AND START THE SIMULATION  *************************

/*

    Ipv4Address remoteHostAddr = internetIpIfaces.GetAddress (1);
    // Install and start applications on UEs and remote host
    uint16_t dlPort = 1234;
    uint16_t ulPort = 2000;
    uint16_t otherPort = 3000;
    ApplicationContainer clientApps;
    ApplicationContainer serverApps;
    for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
    {
      ++ulPort;
      ++otherPort;
      PacketSinkHelper dlPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), dlPort));
      PacketSinkHelper ulPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), ulPort));
      PacketSinkHelper packetSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), otherPort));
      serverApps.Add (dlPacketSinkHelper.Install (ueNodes.Get(u)));
      serverApps.Add (ulPacketSinkHelper.Install (remoteHost));
      serverApps.Add (packetSinkHelper.Install (ueNodes.Get(u)));

      UdpClientHelper dlClient (ueIpIface.GetAddress (u), dlPort);
      dlClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
      dlClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

      UdpClientHelper ulClient (remoteHostAddr, ulPort);
      ulClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
      ulClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

      UdpClientHelper client (ueIpIface.GetAddress (u), otherPort);
      client.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
      client.SetAttribute ("MaxPackets", UintegerValue(1000000));

      clientApps.Add (dlClient.Install (remoteHost));
      clientApps.Add (ulClient.Install (ueNodes.Get(u)));
      if (u+1 < ueNodes.GetN ())
        {
          clientApps.Add (client.Install (ueNodes.Get(u+1)));
        }
      else
        {
          clientApps.Add (client.Install (ueNodes.Get(0)));
        }
    }
    serverApps.Start (Seconds (0.01));
    clientApps.Start (Seconds (0.01));
*/

    // *****************  GET SOME VALUES FROM THE SIMULATION  *************************
    for (uint32_t u = 0; u < numberOfStreamingClients; ++u)
    {
      // get bandwidth values for each UE, from within the running simulation
      Simulator::Schedule ( Seconds(0.5+u), &checkBandwidth , ueLteDevs.Get (u), enbLteDevs.Get (0));

    }

    // reset the contents of our stop file
    Simulator::Schedule ( Seconds(0), &writeSimulationValue );

    // check the file contents and stop the simulator if equal to 1
    for (uint32_t u = 1; u < simTime; ++u)
    {
        // stop the simulator
        Simulator::Schedule ( Seconds(u), &checkSimulation, u );
    }


    // ***************** ANIMATION *****************
    // Create the animation object and configure for specified output
    //AnimationInterface anim (animFile);
    //anim.EnablePacketMetadata (); // Optional
    //anim.EnableIpv4L3ProtocolCounters (Seconds (0), Seconds (10)); // Optional

    // from lena-x2-handover-measures.cc
    // connect custom trace sinks for RRC connection establishment and handover notification
    Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/ConnectionEstablished",
                   MakeCallback (&NotifyConnectionEstablishedEnb));

    // only change direction if we want to use mobility
    if( useMobility){
        Config::Connect ("/NodeList/*/$ns3::MobilityModel/CourseChange",
                   MakeCallback (&CourseChange));
    }

    // *****************  START THE SIMULATION  *************************

    // from D-LiTE
    // these enable some key performance indicators (KPIs)
    // files are created and these are view in the root ns3 folder
    // see src/lte/doc/source/lte-user.rst - line 196
    // lteHelper->EnableMacTraces();
    // lteHelper->EnableRlcTraces();
    // lteHelper->EnablePdcpTraces();

    // if we want to generate traffic..;
    if (trafficGeneration) {

				std::cout << "\n";
        // if we want to generate wireshark pcap files
        if (pcapGenerate) {
            std::cout << "Enable Traces\n";
            // CREATE WIRESHARK TRACES
            p2ph.EnablePcapAll("LENA_P2P_Link");
            csma.EnablePcapAll ("LENA_CSMA_Devices");
        }
        // if we want to generate wireshark pcap files

        // from D-LiTE
        // these enable some key performance indicators (KPIs)
        // files are created and these are view in the root ns3 folder
        // see src/lte/doc/source/lte-user.rst - line 196
        // lteHelper->EnableMacTraces();
        // lteHelper->EnableRlcTraces();
        // lteHelper->EnablePdcpTraces();

        // if we want TCP traffic, if not then UDP traffic
        if(tcp)
        {

			  uint16_t sinkPort = 20000;

				// from ue node 18.0.0.2
			  // Address sinkAddress (InetSocketAddress (ueIpIface.GetAddress (0), sinkPort));
			  // PacketSinkHelper packetSinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), sinkPort));
			  // ApplicationContainer sinkApps = packetSinkHelper.Install (ueNodes.Get (0));

				// from ue node 18.1.0.3
				Address sinkAddress (InetSocketAddress (interfacesUE.GetAddress (1), sinkPort));
			PacketSinkHelper packetSinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), sinkPort));
				// the more UE you add, you need to increase this to numberOfStreamingClients less 1
			ApplicationContainer sinkApps = packetSinkHelper.Install (UE.Get (numberOfStreamingClients-1));


			sinkApps.Start (Seconds (0.0));
			sinkApps.Stop (Seconds (5));

			Ptr<Socket> ns3TcpSocket = Socket::CreateSocket (remoteNodes.Get (1), TcpSocketFactory::GetTypeId ());
			Ptr<MyApp> app = CreateObject<MyApp> ();
			app->Setup (ns3TcpSocket, sinkAddress, 1400, 5000000, DataRate ("1000Mb/s"));

			remoteNodes.Get (1)->AddApplication (app);
			AsciiTraceHelper asciiTraceHelper;
			Ptr<OutputStreamWrapper> stream1 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-window-newreno.txt");
			ns3TcpSocket->TraceConnectWithoutContext ("CongestionWindow", MakeBoundCallback (&CwndChange, stream1));

			Ptr<OutputStreamWrapper> stream4 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-rtt-newreno.txt");
			ns3TcpSocket->TraceConnectWithoutContext ("RTT", MakeBoundCallback (&RttChange, stream4));

			Ptr<OutputStreamWrapper> stream2 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-data-newreno.txt");
			sinkApps.Get(0)->TraceConnectWithoutContext("Rx",MakeBoundCallback (&Rx, stream2));

			app->SetStartTime (Seconds (0.1));
			app->SetStopTime (Seconds (5));

    }
      else
      {

				Ipv4Address remoteHostAddr = internetIpIfaces.GetAddress (1);
        // Install and start applications on UEs and remote host
        uint16_t dlPort = 1234;
        uint16_t ulPort = 2000;
        uint16_t otherPort = 3000;
        ApplicationContainer clientApps;
        ApplicationContainer serverApps;
        for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
        {
          ++ulPort;
          ++otherPort;
          PacketSinkHelper dlPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), dlPort));
          PacketSinkHelper ulPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), ulPort));
          PacketSinkHelper packetSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), otherPort));
          serverApps.Add (dlPacketSinkHelper.Install (UE.Get(u)));
          serverApps.Add (ulPacketSinkHelper.Install (remoteHost));
          serverApps.Add (packetSinkHelper.Install (UE.Get(u)));

          UdpClientHelper dlClient (interfacesUE.GetAddress (1), dlPort);
          dlClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          dlClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

          UdpClientHelper ulClient (remoteHostAddr, ulPort);
          ulClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          ulClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

          UdpClientHelper client (interfacesUE.GetAddress (1), otherPort);
          client.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          client.SetAttribute ("MaxPackets", UintegerValue(1000000));

          clientApps.Add (dlClient.Install (remoteHost));
          clientApps.Add (ulClient.Install (UE.Get(u)));
          if (u+1 < UE.GetN ())
            {
              clientApps.Add (client.Install (UE.Get(u+1)));
            }
          else
            {
              clientApps.Add (client.Install (UE.Get(0)));
            }
        }
        serverApps.Start (Seconds (0.01));
        clientApps.Start (Seconds (0.01));
        }
    }

    std::cout << "Start Simulator\n";
    t_start = std::chrono::high_resolution_clock::now();
    Simulator::Stop(Seconds(simTime));
    Simulator::Run();

    std::cout << "\n";
    std::cout << "Done\n";

    Simulator::Destroy();
    return 0;

}
