 /* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
 /*
 *   Copyright (c) 2011 Centre Tecnologic de Telecomunicacions de Catalunya (CTTC)
 *   Copyright (c) 2015, NYU WIRELESS, Tandon School of Engineering, New York University
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation;
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   Author: Marco Miozzo <marco.miozzo@cttc.es>
 *           Nicola Baldo  <nbaldo@cttc.es>
 *
 *   Modified by: Marco Mezzavilla < mezzavilla@nyu.edu>
 *        	 	  Sourjya Dutta <sdutta@nyu.edu>
 *        	 	  Russell Ford <russell.ford@nyu.edu>
 *        		  Menglei Zhang <menglei@nyu.edu>
 *                Jason Quinlan j.quinlan@cs.ucc.ie
 *                Andrew Nash (MISL Summer of Code 2019)
 *                Kieran Horgan (MISL Summer of Code 2019)
 *
 */

 // sample execution code
 // ./waf --run "scratch/mmwave-simple-epc-tap-nas-v2.2.2 --numberOfNodes=2 --seedValue=1 --mmwaveScheduler=tti --simTime=7.00 --distance1=10.0 --distance2=14.0 --distance3=18.0 --distance4=22.0 --distance5=26.0 --distance6=30.0 --symPerSf=96 --sfPeriod=400 --channel_scenario=RMa --useMobility=false --pcapGenerate=false --tcp=true --trafficGeneration=true"
 
 // or simply - create traffic using TCP and generate pcap files
 
 // ./waf --run "scratch/mmwave-simple-epc-tap-nas-v2.2.2 --pcapGenerate=true --tcp=true --trafficGeneration=true"

// original files
#include "ns3/epc-helper.h"
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "ns3/internet-module.h"
#include "ns3/mobility-module.h"
#include "ns3/applications-module.h"
#include "ns3/point-to-point-helper.h"
#include "ns3/config-store.h"

// mmwave helpers
#include "ns3/mmwave-helper.h"
#include "ns3/mmwave-point-to-point-epc-helper.h"

// new code
#include "ns3/tap-bridge-helper.h"
#include "ns3/csma-helper.h"
#include "ns3/netanim-module.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <iostream>
#include <iomanip>

// added for building propagation
#include <ns3/buildings-helper.h>
#include <ns3/buildings-module.h>

#include "ns3/on-off-helper.h"
#include "ns3/packet-sink-helper.h"

// flow monitoring
#include "ns3/flow-monitor-helper.h"

// real-time versus sim-time
#include <chrono>

using namespace ns3;
//using namespace mmwave;

// real-time variable
std::chrono::system_clock::time_point startTime;

/**
 * Sample simulation script for mmwave+EPC.
 */
NS_LOG_COMPONENT_DEFINE ("MMwaveEpcFirstExample");

// only used for testing with out TapBridge
class MyAppTag : public Tag
{
public:
	MyAppTag ()
	{
	}

	MyAppTag (Time sendTs) : m_sendTs (sendTs)
	{
	}

  static TypeId GetTypeId (void)
  {
    static TypeId tid = TypeId ("ns3::MyAppTag")
      .SetParent<Tag> ()
      .AddConstructor<MyAppTag> ();
    return tid;
  }

  virtual TypeId  GetInstanceTypeId (void) const
  {
    return GetTypeId ();
  }

  virtual void  Serialize (TagBuffer i) const
  {
  	i.WriteU64 (m_sendTs.GetNanoSeconds());
  }

  virtual void  Deserialize (TagBuffer i)
  {
  	m_sendTs = NanoSeconds (i.ReadU64 ());
  }

  virtual uint32_t  GetSerializedSize () const
  {
  	return sizeof (m_sendTs);
  }

  virtual void Print (std::ostream &os) const
  {
  	std::cout << m_sendTs;
  }

	Time m_sendTs;
};


class MyApp : public Application
{
public:

  MyApp ();
  virtual ~MyApp();
  void ChangeDataRate (DataRate rate);
  void Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);



private:
  virtual void StartApplication (void);
  virtual void StopApplication (void);

  void ScheduleTx (void);
  void SendPacket (void);

  Ptr<Socket>     m_socket;
  Address         m_peer;
  uint32_t        m_packetSize;
  uint32_t        m_nPackets;
  DataRate        m_dataRate;
  EventId         m_sendEvent;
  bool            m_running;
  uint32_t        m_packetsSent;
};

MyApp::MyApp ()
  : m_socket (0),
    m_peer (),
    m_packetSize (0),
    m_nPackets (0),
    m_dataRate (0),
    m_sendEvent (),
    m_running (false),
    m_packetsSent (0)
{
}

MyApp::~MyApp()
{
  m_socket = 0;
}

void
MyApp::Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
  m_socket = socket;
  m_peer = address;
  m_packetSize = packetSize;
  m_nPackets = nPackets;
  m_dataRate = dataRate;
}

void
MyApp::ChangeDataRate (DataRate rate)
{
	m_dataRate = rate;
}

void
MyApp::StartApplication (void)
{
  m_running = true;
  m_packetsSent = 0;
  m_socket->Bind ();
  m_socket->Connect (m_peer);
  SendPacket ();
}

void
MyApp::StopApplication (void)
{
  m_running = false;

  if (m_sendEvent.IsRunning ())
    {
      Simulator::Cancel (m_sendEvent);
    }

  if (m_socket)
    {
      m_socket->Close ();
    }
}

void
MyApp::SendPacket (void)
{
	Ptr<Packet> packet = Create<Packet> (m_packetSize);
	MyAppTag tag (Simulator::Now ());

	m_socket->Send (packet);
  	if (++m_packetsSent < m_nPackets)
	{
	    ScheduleTx ();
	}
}



void
MyApp::ScheduleTx (void)
{
	if (m_running)
	{
		Time tNext (Seconds (m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate ())));
		m_sendEvent = Simulator::Schedule (tNext, &MyApp::SendPacket, this);
	}
}

static void
CwndChange (Ptr<OutputStreamWrapper> stream, uint32_t oldCwnd, uint32_t newCwnd)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << oldCwnd << "\t" << newCwnd << std::endl;
}


static void
RttChange (Ptr<OutputStreamWrapper> stream, Time oldRtt, Time newRtt)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << oldRtt.GetSeconds () << "\t" << newRtt.GetSeconds () << std::endl;
}



static void Rx (Ptr<OutputStreamWrapper> stream, Ptr<const Packet> packet, const Address &from)
{
	*stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << packet->GetSize()<< std::endl;
}


/*
* from lena-x2-handover-measures.cc
* this prints out the "eNB CellId", "IMSI" and "RNTI" for each connected UE
* the "eNB CellId" is the ID of the cell tower
*
* International Mobile Subscriber Identity (IMSI)
* the "IMSI" is a number incremental from 1 in the order in which the UE are allocated to the system
*
* Radio Network Temporary Identifier (RNTI)
* the "RNTI" is a random number for each UE, typically in reverse order to which the UE are allocated to the system
* i.e. first allocated has the highest RNTI, but not always
*/
void
NotifyConnectionEstablishedEnb (std::string context,
                                uint64_t imsi,
                                uint16_t cellid,
                                uint16_t rnti)
{

    // this returns the NodeList Number, which is 3 greater than the RNTI
//     std::cout << " " << context << std::endl;
//     context.erase (0,10);
//     context.erase (context.begin()+1, context.end());
//     std::cout << context ", " <<
     std::cout <<  "eNB CellId " << cellid
              << " IMSI " << imsi
              << " RNTI " << rnti
              << std::endl;

    // reduced print output
    // based on (Cell, IMSI, RNTI)
//    std::cout <<  "eNB: " << cellid
//              << ", " << imsi
//              << ", " << rnti
//              << std::endl;
}

// function to stop the simulation, by reading from an input file in which we can change the value from 0 (run) to 1 (stop)
void checkSimulation(double simTime)
{

    // read from the input file
    std::ifstream inputFile("lteStopFile.txt");
    std::string line;

    // if the file exists and is open
    if ( inputFile.is_open() )
    {
        // get the first file from the file
        std::getline(inputFile, line);

        // output a print line for us to review
        std::cout << "At real time " << std::chrono::duration<double, std::milli>(std::chrono::system_clock::now()-startTime).count()/1000    << " s and simulation time " << simTime << " s the content of the file is " << line <<  "\n";

        // if the input from the file is 1
        if ( line == "1" )
        {
            // output a priont line for us to review
            std::cout << "Kill the clients at " << simTime << "\n";
            // stop the simulation
            Simulator::Stop(Simulator::Now());
        }
        // close the file
        inputFile.close();

    }
    // if the file does not exist or can not be opened
    else
    {
        std::cout << "File not opening" <<  "\n";
    }
}

// function to reset the contents of the input file used to stop the simulation
void writeSimulationValue()
{
    // read from the input file
    std::ofstream inputFile("lteStopFile.txt");
    std::string line;

    // if the file exists and is open
    if ( inputFile.is_open() )
    {
        // write a '0' to the input file, by overwriting the existing content
        inputFile << "0";
        // close the file
        inputFile.close();

    }
    // if the file does not exist or can not be opened
    else
    {
        std::cout << "File not opening" <<  "\n";
        return;
    }
}

// function to get the bandwidth and other metrics from the eNB and UE
void checkBandwidth(Ptr<NetDevice> ueDevice, Ptr<NetDevice> enbDevice)
{
    Ptr<MmWaveEnbNetDevice> enbMmWaveDevice = enbDevice->GetObject<MmWaveEnbNetDevice> ();
    Ptr<MmWaveEnbPhy> enbPhy = enbMmWaveDevice->GetPhy ();

    // this works, but need to change it before clients are added
    //enbMmWaveDevice->SetBandwidth(36);

    Ptr<MmWaveUeNetDevice> ueMmWaveDevice = ueDevice->GetObject<MmWaveUeNetDevice> ();
    Ptr<LteUeRrc> ueRrc = ueMmWaveDevice->GetRrc ();
    Ptr<MmWaveUePhy> uePhy = ueMmWaveDevice->GetPhy ();

    std::cout << "\n";
    std::cout << "********\n";
    uint16_t ueCellId = ueRrc->GetCellId ();
    std::cout << "ueCellId " << ueCellId << "\n";
    uint16_t enbCellId = enbMmWaveDevice->GetCellId ();
    std::cout << "enbCellId " << enbCellId << "\n";
    uint8_t ueDlBandwidth = ueRrc->GetDlBandwidth ();
    std::cout << "UE-1 download - # RB = " << (uint32_t) ueDlBandwidth << ", => " << (((uint32_t) ueDlBandwidth) * .2) << "MHz\n";
    uint8_t ueUlBandwidth = ueRrc->GetUlBandwidth ();
    std::cout << "UE-1 upload " << (uint32_t) ueUlBandwidth << "\n";
    uint8_t ueDlEarfcn = ueRrc->GetDlEarfcn ();
    std::cout << "ueDlEarfcn " << (uint32_t) ueDlEarfcn << "\n";
    uint8_t ueUlEarfcn = ueRrc->GetUlEarfcn ();
    std::cout << "ueUlEarfcn " << (uint32_t) ueUlEarfcn << "\n";
    uint64_t ueImsi = ueMmWaveDevice->GetImsi ();
    std::cout << "ueImsi " << ueImsi << "\n";
    double ueTxPower = uePhy->GetTxPower();
    std::cout << "ueTxPower " << ueTxPower << "\n";
    double enbTxPower = enbPhy->GetTxPower();
    std::cout << "enbTxPower " << enbTxPower << "\n";

    // new code
    uint16_t ueRnti = ueRrc->GetRnti ();
    std::cout << "ueRnti " << (uint32_t) ueRnti << "\n";
    // no specific Ul or Dl fnc for Earfnc, only getearfnc
    uint8_t enbDlEarfcn = enbMmWaveDevice->GetEarfcn ();
    std::cout << "enbEarfcn " << (uint32_t) enbDlEarfcn << "\n";
    //
    uint8_t enbBandwidth = enbMmWaveDevice->GetBandwidth ();
    std::cout << "enbBandwidth " << (uint32_t) enbBandwidth << "\n";

}

// change the direction of the UE
static void CourseChange(std::string foo, Ptr<const MobilityModel> mobility)
{
    // get the current simulation time
    Time now = Simulator::Now();
    // get the position of the UE
    Vector pos = mobility->GetPosition();

    // this returns the NodeList Number, which is 3 greater than the RNTI
    // std::cout << " " << foo << std::endl;
    foo.erase (0,10); // remove "/NodeList/" from start of string
    foo.erase (foo.begin()+1, foo.end()); // remvoe remainder of the string
    int value = atoi(foo.c_str()); // convert string nubmer to int
    value=value-3; // reduce int value by 3 to get RNTI number of the UE

    // print the co-ordinates of the UE
    // std::cout <<"Time: "<<now.GetSeconds()<<" RNTI: "<<value<<" POS: x="<<pos.x<< ", y="<<pos.y<<", z="<<pos.z<<std::endl;

    // less print out using (Time, RNTI, xPOs, yPOS, zPOS)
    // lets also format the output, so info in in columns and easier to read -
    //      - using std::setw(2) - mandate two spaces for output
    //      - using std::setprecision(6) - mandate 6 character precision including decimal points
    std::cout <<"Time: "<< std::setw(6) << now.GetSeconds()<<", "<<value<<", " << std::setw(9)  <<pos.x<< ", "<< std::setw(9) <<pos.y<<", "<<pos.z<<std::endl;
}


// define the structure of the LENA+mmwave architecture
int
main (int argc, char *argv[])
{

    // good papers on 5G mmwave
    // https://arxiv.org/pdf/1703.08985.pdf - TCP in 5G mmWave Networks: Link Level Retransmissions and MP-TCP
    // http://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=5876482 - mmWave Mobile Broadband (MMB): Unleashing the 3-300GHz Spectrum ([4] from paper above)


    // *****************  LETS DEFINE SOME VARIABLES ************************
    // flowMonitoring for PacketFlows
    bool flowMonitor = false;
    // use tcp
	  bool tcp = false;
	  // generate pcap
	  bool pcapGenerate = false;
	  // switch on traffic generation
	  bool trafficGeneration = false;

    uint16_t numberOfeNodeBs = 1;
    uint16_t numberOfNodes = 2;
    double simTime = 2.0;
    double interPacketInterval = 100;  // 500 microseconds
    double minDistance = 10.0;  // eNB-UE distance in meters
    double maxDistance = 150.0;  // eNB-UE distance in meters
    // - Hybrid Automatic Repeat reQuest - mac layer - used for adding redundancy to
    // lost packets to improve retrieval of correct version - default is true for MmWaveHelper
    // modified 28-03-2019 - changed to false
    bool harqEnabled = false; // true;
    // - Radio Link Control - acknowledgment mode -
    // used for retransmitting lost packets - default is false
    bool rlcAmEnabled = true; // false;
    // - Transmission Time Interval (TTI):

    // fixedTti=true - fixed number of symbols based on "SymbolPerSlot × SymbolLength"
        // SymbolPerSlot defined in MmWavePhyMacCommon =
        // SymbolPerSlot can also be set per scheduler - default value of 6 in all four schedulers
        // but will result in only one UE being scheduled per subframe, which would be highly inefficient in a multi-user cell
        // SymbolLength defined as "m_symbolPeriod" in MmWavePhyMacCommon

    // fixedTti=true - variable number of symbols based on length of packet or
    // transport block (TB) - default is false
    bool fixedTti = false;
    // OFDM symbols per subframe - default 24
    unsigned symPerSf = 24; // 24;
    // Subframe period = 4.16 * symPerSf - default is 100
    double sfPeriod = 100.0; // 100.0;
    // Number of concurrent stop-and-wait Hybrid ARQ processes per user - default is 100.0
    // updated 28-03-2019 - from 200 to 400
    double TbDecodeLatency = 400.0; // 200.0;
    // run value for ns3, so as to mandate the same output everytime
    unsigned run = 1;
    // Enable small scale fading for beam forming
    bool smallScale = true; // true;
    // Set a fixed speed (even if constant position) so doppler > 0 for testing for beam forming
    bool beam_fixSpeed = false; // true;
    // UE speed (m/s) for fixed speed test for beam forming
    double speed = 0.0; // 3;

    // new code - my variables - all default values
    uint16_t numberOfStreamingClients = 6;
    bool backgroundTraffic = false;
    uint32_t bandwidth = 6;
    //double interPacketInterval = 1;  // .5 microseconds
    // original LTE distances
    // double distance[12] = {50, 100, 150, 200, 250, 300, 350, 400, 450, 500, 550, 600};
    // distances defined from MmWave distances above - distance of 28 per ue
    double distance[6] = {10, 38, 66, 94, 122, 150};
    uint16_t numRB = 25;
    uint16_t seedValue = 1;
    double mmwaveTxPower = 25.0;
    std::string mmwaveScheduler = "tti";
    std::string mmwavePathloss = "prop";
    bool useMobility = false;
    bool usePrintMobility = false;
    std::string device_DataRate = "100Gb/s";
    bool useFading = false;
    bool mobileFading = true;
    bool vehicleFading = false;

    // Name of NetAnim file for animation output
    std::string animFile = "ns3-lte-animation.xml";

    // previous static values converted to variables
    double channel_Delay = 0.001;
    // CQI threshold for schedulers
    uint32_t CQI_timer_threshold = 1000;
    // Number of symbols per slot in Fixed TTI mode for schedulers
    uint8_t SymPerSlot = 100;
    double hBS = 0.0; //base station antenna height in meters;
    double hUT = 0.0; //user antenna height in meters;

    // enable the 3gpp channel
    bool channel3GPP = false;
    // set channel condition - "l" - los, "n" - nlos and "a" - all
    std::string channel_condition = "a";
    // set channel scenario - RMa: Rural Macro, UMa: Urban Macro, UMi-StreetCanyon: Urban Micro,
    // InH-OfficeMixed: Indoor Office Mixed, InH-OfficeOpen: Indoor Office Open,
    // InH-ShoppingMall: Indoor Shopping Mall
    std::string channel_scenario = "RMa";
    // set nlos propogation model
    bool channel_nlos = false;
    // set shadowing
    bool channel_shadowing = true;
    // update los/nlos condition when ue moves
    bool update_channel_condition = true;
    // interval after which the channel for a moving user is updated
    uint16_t channel3GPP_updatePeriod = 100;
    // Set true to use cell scanning method, false to use the default power method.
    bool channel3GPP_cellScan = false;
    // use blockage or not
    bool channel3GPP_blockage = true;
    // use blockage model with UT in portrait mode
    bool channel3GPP_portraitMode = true;
    // number of non-self blocking obstacles
    uint16_t channel3GPP_numNonselfBlocking = 4;

    // enable the Ray tracing channel
    bool channelRaytracing = false;
    // start point of the simulation, the range of data point is [0, 260] meters
    uint16_t raytrace_startDistance = 0;
    // The speed of the node (m/s)
    double raytrace_speed = 1.0;

    // enable the BeamForming channel
    bool channelBeamForming = false;
    // Time (ms) between periodic updating of channel matrix/beamforming vectors
    double beam_longTermUpdatePeriod = 100.0;

    // building propagation loss variables
    bool enableBuildings = false;
    // channel state - "'l' for LOS, 'n' for NLOS, 'o' for outage, 'a' for all"
    std::string channel_state = "a";
    // reducing the tx power by this amount
    double loss_Fixed_Db = 0.0;
    // boolean to mandate you reduce the tx power by the fixed amount above
	bool fixed_Loss_Tst = false;

    // building values - all default values
    double xMin = 20.0;
    double xMax = 40.0;
    double yMin = 0.0;
    double yMax = 20.0;
    double zMin = 0.0;
    double zMax = 20.0;
    // The type of building - default Residential
    std::string building_type = "Residential";
    // Type of material of which the external walls are made - default ConcreteWithWindows
    std::string extWalls_type = "ConcreteWithWindows";
    uint32_t num_Floors = 1; // The number of floors in this building - default uint32_t 1
    uint32_t num_RoomsX = 1; // The number of rooms in the X axis - default uint32_t 1
    uint32_t num_RoomsY = 1; // The number of rooms in the Y axis - default uint32_t 1

    // LteEnb values
    // The interval for sending system information (Time value) - default 80
    double lteEnb_SystemInformationPeriodicity = 5.0; // 80.0; // 5.0;
    // The SRS periodicity in milliseconds - default 40
    uint32_t lteEnb_SrsPeriodicity = 40; // 40; // 320;
    // Time in ms of initial SIB message - default 16
    uint32_t lteEnb_FirstSibTime = 2; // 16; // 2;

    // LteRlc values
    // How much to wait to issue a new Report Buffer Status since the last time - default 20.0
    double LteRlc_reportBufferStatusTimer = 20; // 100.0;

    // Command line arguments
    CommandLine cmd;
    cmd.AddValue ("writeFlowMonitor", "<0/1> to enable Flow Monitor and write their results", flowMonitor);
    cmd.AddValue ("tcp", "<true> to enable UDP and <false> to enable UDP", tcp);
    cmd.AddValue ("pcapGenerate", "<true> to enable wireshark pcap files", pcapGenerate);
    cmd.AddValue ("trafficGeneration", "<true> to generate traffic", trafficGeneration);
    cmd.AddValue("numberOfeNodeBs", "Number of eNBs", numberOfeNodeBs);
    cmd.AddValue("numberOfNodes", "Number of UEs per eNB", numberOfNodes);
    cmd.AddValue("simTime", "Total duration of the simulation [s])", simTime);
    cmd.AddValue("interPacketInterval", "Inter-packet interval [us])", interPacketInterval);
    cmd.AddValue("harq", "Enable Hybrid ARQ", harqEnabled);
    cmd.AddValue("rlcAm", "Enable RLC-AM", rlcAmEnabled);
    cmd.AddValue("symPerSf", "OFDM symbols per subframe", symPerSf);
    cmd.AddValue("sfPeriod", "Subframe period = 4.16 * symPerSf", sfPeriod);
    cmd.AddValue("TbDecodeLatency", "Number of concurrent stop-and-wait Hybrid ARQ processes per user", TbDecodeLatency);
    cmd.AddValue("fixedTti", "Fixed TTI scheduler", fixedTti);
    // will assume "run" should be passed here
    cmd.AddValue("run", "run for RNG (for generating different deterministic sequences for different drops)", run);
    cmd.AddValue("minDistance", "min eNB-UE distance in meters", minDistance);
    cmd.AddValue("maxDistance", "max eNB-UE distance in meters", maxDistance);
    cmd.AddValue("beam_longTermUpdatePeriod", "Long Term Update Period for beam forming", beam_longTermUpdatePeriod);
    cmd.AddValue("smallScale", "Small Scale Fading for beam forming", smallScale);
    cmd.AddValue("beam_fixSpeed", "Fix Speed for beam forming", beam_fixSpeed);
    cmd.AddValue("speed", "UE speed for beam forming", speed);

    // my command line arguments
    cmd.AddValue("backgroundTraffic", "", backgroundTraffic);
    cmd.AddValue("numberOfStreamingClients", "Number of Streaming UEs per eNB", numberOfStreamingClients);
    cmd.AddValue("distance1", "Distance between eNBs [m]", distance[0]);
    cmd.AddValue("distance2", "Distance between eNBs [m]", distance[1]);
    cmd.AddValue("distance3", "Distance between eNBs [m]", distance[2]);
    cmd.AddValue("distance4", "Distance between eNBs [m]", distance[3]);
    cmd.AddValue("distance5", "Distance between eNBs [m]", distance[4]);
    cmd.AddValue("distance6", "Distance between eNBs [m]", distance[5]);
    cmd.AddValue("numRB", "Number of Resource Blocks", numRB);
    cmd.AddValue("seedValue", "Seed Value for NS3", seedValue);
    cmd.AddValue("mmwaveTxPower", "UE/eNB tx power", mmwaveTxPower);
    cmd.AddValue("bandwidth", "Bandwidth level to use", bandwidth);
    cmd.AddValue("mmwaveScheduler", "MmWave Scheduler Model", mmwaveScheduler);
    cmd.AddValue("mmwavePathloss", "MmWave Pathloss Model", mmwavePathloss);
    cmd.AddValue("useMobility", "Define a Mobility model for the UE", useMobility);
    cmd.AddValue("device_DataRate", "Data Rate per Link", device_DataRate);
    cmd.AddValue("channel_Delay", "Channel Delay per link", channel_Delay);
    cmd.AddValue("useFading", "If Fading should be used", useFading);
    cmd.AddValue("mobileFading", "Use Mobile Fading model", mobileFading);
    cmd.AddValue("vehicleFading", "Use Vehicle Fading model", vehicleFading);
    cmd.AddValue("animFile", "Define the NetANmi animation file name", animFile);
    cmd.AddValue("CQI_timer_threshold", "CQI threshold for schedulers", CQI_timer_threshold);
    cmd.AddValue("SymPerSlot", "Number of symbols per slot in Fixed TTI mode for schedulers", SymPerSlot);


    // additional values added for channel
    cmd.AddValue("hBS", "base station antenna height in meters", hBS);
    cmd.AddValue("hUT", "user antenna height in meters", hUT);
    cmd.AddValue("channel3GPP", "Use a 3GPP channel model", channel3GPP);
    cmd.AddValue("channel_condition", "set channel condition - l - los, n - nlos and a - allset ", channel_condition);
    cmd.AddValue("channel_scenario", "channel scenario - RMa: Rural Macro, UMa: Urban Macro, UMi-StreetCanyon: Urban Micro, InH-OfficeMixed: Indoor Office Mixed, InH-OfficeOpen: Indoor Office Open, InH-ShoppingMall: Indoor Shopping Mall", channel_scenario);
    cmd.AddValue("channel_nlos", "set nlos propogation model", channel_nlos);
    cmd.AddValue("channel_shadowing", "set shadowing", channel_shadowing);
    cmd.AddValue("update_channel_condition", "update los/nlos condition when ue move", update_channel_condition);
    cmd.AddValue("channel_state", "channel state - l for LOS, n for NLOS, o for outage, a for all", channel_state);
    cmd.AddValue("loss_Fixed_Db", "reducing the tx power by this amount ", loss_Fixed_Db);
    cmd.AddValue("fixed_Loss_Tst", "boolean to reduce the tx power by the fixed amount above", fixed_Loss_Tst);
    cmd.AddValue("channel3GPP_updatePeriod", "interval after which the channel for a moving user is updated", channel3GPP_updatePeriod);
    cmd.AddValue("channel3GPP_cellScan", "Set true to use cell scanning method, false to use the default power method.", channel3GPP_cellScan);
    cmd.AddValue("channel3GPP_blockage", "use blockage or not", channel3GPP_blockage);
    cmd.AddValue("channel3GPP_portraitMode", "use blockage model with UT in portrait mode", channel3GPP_portraitMode);
    cmd.AddValue("channel3GPP_numNonselfBlocking", "number of non-self blocking obstacles", channel3GPP_numNonselfBlocking);


    // additional values added for ray tracing channel
    cmd.AddValue("channelRaytracing", "enable the Ray tracing channel", channelRaytracing);
    cmd.AddValue("raytrace_startDistance", "start point of the simulation, the range of data point is [0, 260] meters", raytrace_startDistance);
    cmd.AddValue("raytrace_speed", "The speed of the node (m/s)", raytrace_speed);

    // additional values added for Beam Forming channel
    cmd.AddValue("channelBeamForming", "enable the Beam Forming channel", channelBeamForming);

    // additional values added for buildings
    cmd.AddValue("enableBuildings", "enable the building propagation loss variables", enableBuildings);
    cmd.AddValue("xMin", "building values", xMin);
    cmd.AddValue("xMax", "building values", xMax);
    cmd.AddValue("yMin", "building values", yMin);
    cmd.AddValue("yMax", "building values", yMax);
    cmd.AddValue("zMin", "building values", zMin);
    cmd.AddValue("zMax", "building values", zMax);
    cmd.AddValue("building_type", "the type of building - default Residential", building_type);
    cmd.AddValue("extWalls_type", "Type of material of which the external walls are made - default ConcreteWithWindows", extWalls_type);
    cmd.AddValue("num_Floors", "The number of floors in this building - default uint32_t 1", num_Floors);
    cmd.AddValue("num_RoomsX", "The number of rooms in the X axis - default uint32_t 1", num_RoomsX);
    cmd.AddValue("num_RoomsY", "The number of rooms in the Y axis - default uint32_t 1", num_RoomsY);

    // LteEnb values
    cmd.AddValue("lteEnb_SystemInformationPeriodicity", "The interval for sending system information (Time value)", lteEnb_SystemInformationPeriodicity);
    cmd.AddValue("lteEnb_SrsPeriodicity", "The SRS periodicity in milliseconds", lteEnb_SrsPeriodicity);
    cmd.AddValue("lteEnb_FirstSibTime", "Time in ms of initial SIB message", lteEnb_FirstSibTime);

    // LteRlc values
    cmd.AddValue("LteRlc_reportBufferStatusTimer", "How much to wait to issue a new Report Buffer Status since the last time", LteRlc_reportBufferStatusTimer);

    // parse the input
    cmd.Parse(argc, argv);

    // original code - how much of this do we actually need

    int PacketSize = 2500;
	Config::SetDefault ("ns3::TcpSocketBase::MinRto", TimeValue (MilliSeconds (200)));
	Config::SetDefault ("ns3::Ipv4L3Protocol::FragmentExpirationTimeout", TimeValue (Seconds (1)));
	Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (PacketSize));
	Config::SetDefault ("ns3::TcpSocket::DelAckCount", UintegerValue (1));
	//Config::SetDefault ("ns3::TcpSocket::SndBufSize", UintegerValue (131072*200));
	//Config::SetDefault ("ns3::TcpSocket::RcvBufSize", UintegerValue (131072*200));

	Config::SetDefault ("ns3::TcpSocket::SndBufSize", UintegerValue (131072*180*4));
	Config::SetDefault ("ns3::TcpSocket::RcvBufSize", UintegerValue (131072*180*2));

    //Config::SetDefault ("ns3::Queue::MaxPackets", UintegerValue (100*1000));

	//Config::SetDefault ("ns3::CoDelQueueDisc::Mode", StringValue ("QUEUE_MODE_PACKETS"));
    //Config::SetDefault ("ns3::CoDelQueueDisc::MaxPackets", UintegerValue (50000));
	Config::SetDefault ("ns3::MmWavePhyMacCommon::NumHarqProcess", UintegerValue(100));
/*
	Config::SetDefault ("ns3::LteRlcAm::PollRetransmitTimer", TimeValue(MilliSeconds(2.0)));
	Config::SetDefault ("ns3::LteRlcAm::ReorderingTimer", TimeValue(MilliSeconds(1.0)));
	Config::SetDefault ("ns3::LteRlcAm::StatusProhibitTimer", TimeValue(MilliSeconds(1.0)));
	Config::SetDefault ("ns3::LteRlcAm::ReportBufferStatusTimer", TimeValue(MilliSeconds(2.0)));
	Config::SetDefault ("ns3::LteRlcAm::MaxTxBufferSize", UintegerValue (1000 *1000 *10));
*/

    // ***** MMWAVE HELPER VARIABLES *****
    // used to set the Radio Link Control - Enable RLC Acknowledged Mode - default false
    Config::SetDefault ("ns3::MmWaveHelper::RlcAmEnabled", BooleanValue(rlcAmEnabled));
    // used to set the Hybrid Automatic Repeat reQuest - mac layer - acknowledgment mode in mmwave helper - Enable Hybrid ARQ - default true
    Config::SetDefault ("ns3::MmWaveHelper::HarqEnabled", BooleanValue(harqEnabled));

    // ***** MMWAVE PHYSICAL MAC COMMON VARIABLES *****
    // OFDM symbols per subframe - default 24
    Config::SetDefault ("ns3::MmWavePhyMacCommon::SymbolsPerSubframe", UintegerValue(symPerSf));
    // Symbol period in microseconds - default 100.0
    Config::SetDefault ("ns3::MmWavePhyMacCommon::SubframePeriod", DoubleValue(sfPeriod));
    // Number of concurrent stop-and-wait Hybrid ARQ processes per user - default 100.0
    Config::SetDefault ("ns3::MmWavePhyMacCommon::TbDecodeLatency", UintegerValue(TbDecodeLatency));

    // ***** LTE ENB RRC VARIABLES *****
    // The interval for sending system information (Time value) - default double 80
    Config::SetDefault ("ns3::LteEnbRrc::SystemInformationPeriodicity", TimeValue (MilliSeconds (lteEnb_SystemInformationPeriodicity)));
    // The SRS periodicity in milliseconds - default uint32_t 40
    Config::SetDefault ("ns3::LteEnbRrc::SrsPeriodicity", UintegerValue (lteEnb_SrsPeriodicity));
    // Time in ms of initial SIB message - default uint32_t 16
    Config::SetDefault ("ns3::LteEnbRrc::FirstSibTime", UintegerValue (lteEnb_FirstSibTime));
/*
    // ***** LTE RLC AM VARIABLES *****
    // How much to wait to issue a new Report Buffer Status since the last time a new SDU was received - default 20
    Config::SetDefault ("ns3::LteRlcAm::ReportBufferStatusTimer", TimeValue(MicroSeconds(lteEnb_SrsPeriodicity)));

    // ***** LTE RLC UM LOW LAT VARIABLES *****
    // How much to wait to issue a new Report Buffer Status since the last time - default 20.0
    Config::SetDefault ("ns3::LteRlcUmLowLat::ReportBufferStatusTimer", TimeValue(MicroSeconds(lteEnb_SrsPeriodicity)));
*/
    // ***** ENB and UE TXpower VARIABLES *****
    // keep this code commented for now
    // these work - but let's use default values before we move to our own defined values
    // change the tx power for the UE devices
    Config::SetDefault ("ns3::MmWaveUePhy::TxPower", DoubleValue (mmwaveTxPower));
    // change the tx power for the ENB devices
    Config::SetDefault ("ns3::MmWaveEnbPhy::TxPower", DoubleValue (mmwaveTxPower));

    // these do not work - hmmm....
    //Config::SetDefault ("ns3::MmWaveChannelMatrix::NumResourceBlocks", UintegerValue (12));
    //Config::SetDefault ("ns3::MmWavePhyMacCommon::ChunkPerRB", UintegerValue(24));


    // ***********  LETS IMPLEMENT REAL-TIME SIMULATION ***********
    // from 'lena-simple-tap.cc'
    GlobalValue::Bind ("SimulatorImplementationType",
                     StringValue ("ns3::RealtimeSimulatorImpl"));
    GlobalValue::Bind ("ChecksumEnabled", BooleanValue (true));

    // set a seed and run value, based on passed in parameters
    std::cout << "Seed value of " << seedValue << "\n";
    RngSeedManager::SetSeed (seedValue);
    std::cout << "Run value for RNG of " << run << "\n";
    RngSeedManager::SetRun (run);

    // ***********  LETS CREATE SOME ADDITIONAL LOG STATEMENTS ***********
    // Enable Logging - must be run with debug version of ns3

    //LogLevel logLevel = (LogLevel)(LOG_PREFIX_FUNC | LOG_PREFIX_TIME | LOG_LEVEL_ALL);


    //LogComponentEnable ("EpcSgwPgwApplication", LOG_LEVEL_ALL);

    /*
    LogComponentEnable ("LteUeRrc", LOG_LEVEL_ALL);
    EpcSgwPgwApplication
    LogComponentEnable ("LteEnbRrc", LOG_LEVEL_ALL);
    LogComponentEnable("MmWavePointToPointEpcHelper",LOG_LEVEL_ALL);
    LogComponentEnable("EpcUeNas",LOG_LEVEL_ALL);
    LogComponentEnable ("MmWaveSpectrumPhy", LOG_LEVEL_LOGIC);
    LogComponentEnable ("MmWaveUePhy", LOG_LEVEL_DEBUG);
    LogComponentEnable ("MmWaveEnbPhy", LOG_LEVEL_DEBUG);
    LogComponentEnable ("MmWaveUeMac", LOG_LEVEL_LOGIC);
    LogComponentEnable ("UdpClient", LOG_LEVEL_INFO);
    LogComponentEnable ("PacketSink", LOG_LEVEL_INFO);
    LogComponentEnable("PropagationLossModel",LOG_LEVEL_ALL);
    LogComponentEnable("MmWaveFlexTtiMacScheduler",logLevel);
    */

    std::cout << "**** SETUP ****\n";
    std::cout << "MMWAVE and EPC Creation\n";
    Ptr<MmWaveHelper> mmwaveHelper = CreateObject<MmWaveHelper> ();
    Ptr<MmWavePointToPointEpcHelper>  epcHelper = CreateObject<MmWavePointToPointEpcHelper> ();
    mmwaveHelper->SetEpcHelper (epcHelper);
    mmwaveHelper->SetHarqEnabled (harqEnabled);

    // #################  SCHEDULER ###################
    // we need to understand each of these schedulers
	std::cout << "MMWAVE Scheduler - ";

    if (mmwaveScheduler == "ttipf" ){

        // ***** MmWaveFlexTtiPfMac SCHEDULER VARIABLES *****
/*
        // for setting harq - Activate/Deactivate the HARQ [by default is active]
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiPfMacScheduler::HarqEnabled", BooleanValue(harqEnabled));

        // setting channel quality information (CQI) timer threshold - The number of TTIs a CQI is valid (default 1000 - 1 sec.) - meaured in microseconds (us)
        // default value is "100"
        Config::SetDefault ("ns3::MmWaveFlexTtiPfMacScheduler::CqiTimerThreshold", UintegerValue(CQI_timer_threshold));

        // for setting tti - fixed or variable Fix slot size
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiPfMacScheduler::FixedTti", BooleanValue(fixedTti));
*/
        // this is a Proportional Fair (PF) scheduler
        std::cout << "MmWaveFlexTtiPfMacScheduler is now active\n";
        mmwaveHelper->SetSchedulerType("ns3::MmWaveFlexTtiPfMacScheduler");

    }
    else if (mmwaveScheduler == "ttimaxWeight" ){

        // ***** MmWaveFlexTtiMaxWeightMac SCHEDULER VARIABLES *****
/*
        // for setting harq - Activate/Deactivate the HARQ [by default is active]
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxWeightMacScheduler::HarqEnabled", BooleanValue(harqEnabled));

        // setting channel quality information (CQI) timer threshold - The number of TTIs a CQI is valid (default 1000 - 1 sec.) - meaured in microseconds (us)
        // default value is "100"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxWeightMacScheduler::CqiTimerThreshold", UintegerValue(CQI_timer_threshold));

        // for setting tti - fixed or variable Fix slot size
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxWeightMacScheduler::FixedTti", BooleanValue(fixedTti));

        // for setting number of symbols per slot - in all schedulers the
        // default value is 6
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxWeightMacScheduler::SymPerSlot", UintegerValue(SymPerSlot));
*/
        // priority queue-based policy that weighs flows by their relative deadlines for packet delivery
        std::cout << "MmWaveFlexTtiMaxWeightMacScheduler is now active\n";
        mmwaveHelper->SetSchedulerType("ns3::MmWaveFlexTtiMaxWeightMacScheduler");

    }
    else if (mmwaveScheduler == "ttimaxRate" ){

        // ***** MmWaveFlexTtiMaxRateMac SCHEDULER VARIABLES *****
/*
        // for setting harq - Activate/Deactivate the HARQ [by default is active]
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxRateMacScheduler::HarqEnabled", BooleanValue(harqEnabled));

        // setting channel quality information (CQI) timer threshold - The number of TTIs a CQI is valid (default 1000 - 1 sec.) - meaured in microseconds (us)
        // default value is "100"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxRateMacScheduler::CqiTimerThreshold", UintegerValue(CQI_timer_threshold));

        // for setting tti - fixed or variable Fix slot size
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMaxRateMacScheduler::FixedTti", BooleanValue(fixedTti));
*/

        // schedules only the users with the highest SINRs to maximize cell throughput
        std::cout << "MmWaveFlexTtiMaxRateMacScheduler is now active\n";
        mmwaveHelper->SetSchedulerType("ns3::MmWaveFlexTtiMaxRateMacScheduler");

    }
    else if (mmwaveScheduler == "tti" ){

        // ***** MmWaveFlexTtiMac SCHEDULER VARIABLES *****
/*
        // taken from above
        // for setting harq - Activate/Deactivate the HARQ [by default is active]
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMacScheduler::HarqEnabled", BooleanValue(harqEnabled));

        // setting channel quality information (CQI) timer threshold - The number of TTIs a CQI is valid (default 1000 - 1 sec.) - meaured in microseconds (us)
        // default value is "100"
        Config::SetDefault ("ns3::MmWaveFlexTtiMacScheduler::CqiTimerThreshold", UintegerValue(CQI_timer_threshold));

        // for setting tti - fixed or variable Fix slot size
        // default value is "false"
        Config::SetDefault ("ns3::MmWaveFlexTtiMacScheduler::FixedTti", BooleanValue(fixedTti));
*/
        // MmWaveFlexTtiMacScheduler by default - defined in "mmwave-simple-epc.cc"
        // this is a round robin (RR) scheduler
        std::cout << "MmWaveFlexTtiMacScheduler is now active\n";
        mmwaveHelper->SetSchedulerType ("ns3::MmWaveFlexTtiMacScheduler");

        // for setting number of symbols per slot - in all schedulers the
        // default value is 6
        Config::SetDefault ("ns3::MmWaveFlexTtiMacScheduler::SymPerSlot", UintegerValue(SymPerSlot));

    }
    else {

        std::cout << "Default - MmWaveFlexTtiMacScheduler - is now active\n";

    }

    // ################# RESOURCE BLOCK ALLOCATION ###################
    // std::cout << "Bandwidth value of " << bandwidth << "\n";
    // bandwidth seems statically defined at 4 RB, with 28 symbols, giving 72 subbands
    // for now, do not change these values


    // ################# PATHLOSS MODEL ###################
    // we need to understand each of these pathloss models
    std::cout << "MMWAVE Path Loss Model - ";

	if (mmwavePathloss == "buildingObsProp" ){

        std::cout << "BuildingsObstaclePropagationLossModel is now active\n";
        mmwaveHelper->SetAttribute ("PathlossModel", StringValue ("ns3::BuildingsObstaclePropagationLossModel"));
        // enable the building propagation
        enableBuildings = true;
        // taken from "mmwave-simple-building-obstacle.cc"
        mmwaveHelper->Initialize();
    }
    else if (mmwavePathloss == "3gppBuildingProp" ){

        std::cout << "MmWave3gppBuildingsPropagationLossModel is now active\n";
/*
        // enable or disable the LOS/NLOS update when the UE moves
        Config::SetDefault ("ns3::MmWave3gppBuildingsPropagationLossModel::UpdateCondition", BooleanValue(update_channel_condition));

        // we need to define these for MmWave3gppPropagationLossModel to use here
        // update nlos/los condition when ue moves
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::ChannelCondition", StringValue(channel_condition));
        // define the scenario to use
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::Scenario", StringValue(channel_scenario));
        // use the nlos propogation model
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::OptionalNlos", BooleanValue(channel_nlos));
        // enable shadowing
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::Shadowing", BooleanValue(channel_shadowing));

*/
        mmwaveHelper->SetAttribute ("PathlossModel", StringValue ("ns3::MmWave3gppBuildingsPropagationLossModel"));
        // enable the building propagation
        enableBuildings = true;
        // enable 3GPP channel
        channel3GPP = true;

    }
    else if (mmwavePathloss == "3gppProp" ){

        std::cout << "MmWave3gppPropagationLossModel is now active\n";

        // update nlos/los condition when ue moves
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::ChannelCondition", StringValue(channel_condition));
        // define the scenario to use
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::Scenario", StringValue(channel_scenario));
        // use the nlos propogation model
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::OptionalNlos", BooleanValue(channel_nlos));
        // enable shadowing
        Config::SetDefault ("ns3::MmWave3gppPropagationLossModel::Shadowing", BooleanValue(channel_shadowing));

        mmwaveHelper->SetAttribute ("PathlossModel", StringValue  ("ns3::MmWave3gppPropagationLossModel"));

        // enable 3GPP channel
        channel3GPP = true;

    }
    else if (mmwavePathloss == "prop" ){

        // MmWavePropagationLossModel by default
        std::cout << "MmWavePropagationLossModel is now active\n";
/*
        // define the channel state
        Config::SetDefault ("ns3::MmWavePropagationLossModel::ChannelStates", StringValue (channel_state));
        // define if you should use a fixed tx power loss rate
	    Config::SetDefault ("ns3::MmWavePropagationLossModel::FixedLossTst", BooleanValue (fixed_Loss_Tst));
        // value of fixed tx power loss rate to use
	    Config::SetDefault ("ns3::MmWavePropagationLossModel::LossFixedDb", DoubleValue (loss_Fixed_Db));
*/
        mmwaveHelper->SetAttribute ("PathlossModel", StringValue ("ns3::MmWavePropagationLossModel"));

    }
    else {

        std::cout << "Default - MmWavePropagationLossModel - is now active\n";

    }

    // taken from "mmwave-3gpp-channel-example.cc" - may or may not be needed
    //mmwaveHelper->Initialize();


    // #################  FADING MODEL ###################
    // Fading uses the channel models to creatre small-scale fading
    std::cout << "Channel Model - ";

    // if we are using 3GPP pathloss, then use the 3GPP channel and set up UE/eNB heights
    if ( channel3GPP ) {

        // this is real-time, with no traffic travelling over the network
        std::cout << "MmWave3gppChannel is now active\n";
/*
        // interval after which the channel for a moving user is updated, with spatial
        // consistency procedure. If 0, spatial consistency is not used
        Config::SetDefault ("ns3::MmWave3gppChannel::UpdatePeriod", TimeValue(MilliSeconds(channel3GPP_updatePeriod)));
        // Set true to use cell scanning method, false to use the default power method.
        Config::SetDefault ("ns3::MmWave3gppChannel::CellScan", BooleanValue(channel3GPP_cellScan));
        // use blockage or not
        Config::SetDefault ("ns3::MmWave3gppChannel::Blockage", BooleanValue(channel3GPP_blockage));
        // use blockage model with UT in portrait mode
        Config::SetDefault ("ns3::MmWave3gppChannel::PortraitMode", BooleanValue(channel3GPP_portraitMode));
        // number of non-self blocking obstacles
        Config::SetDefault ("ns3::MmWave3gppChannel::NumNonselfBlocking", IntegerValue(channel3GPP_numNonselfBlocking));
*/
        // taken from "mmwave-3gpp-channel-example.cc" - only used with 3gpp models
        mmwaveHelper->SetAttribute ("ChannelModel", StringValue ("ns3::MmWave3gppChannel"));

        // set the height of the UE and eNB, as required by MmWave3gppChannel
        if(channel_scenario.compare("RMa")==0)
        {
            hBS = 35;
            hUT = 1.5;
        }
        else if(channel_scenario.compare("UMa")==0)
        {
            hBS = 25;
            hUT = 1.5;
        }
        else if (channel_scenario.compare("UMi-StreetCanyon")==0)
        {
            hBS = 10;
            hUT = 1.5;
        }
        else if (channel_scenario.compare("InH-OfficeMixed")==0 || channel_scenario.compare("InH-OfficeOpen")==0 || channel_scenario.compare("InH-ShoppingMall")==0)
        {
            hBS = 3;
            hUT = 1;
        }
    }
    else if ( channelRaytracing ) {

        // this is not real-time, even with no traffic traversing it, even with default values
        std::cout << "MmWaveChannelRaytracing is now active\n";
/*
        // implement the ray-tracing channel - taken from ""mmwave-tcp-raytracing-example.cc

        //  start point of the simulation, the range of data point is [0, 260] meters
        Config::SetDefault ("ns3::MmWaveChannelRaytracing::StartDistance", UintegerValue (raytrace_startDistance));
        // The speed of the node (m/s)
        Config::SetDefault ("ns3::MmWaveChannelRaytracing::Speed", DoubleValue (raytrace_speed));
*/
        // use the MmWaveChannelRaytracing channel model
        mmwaveHelper->SetAttribute ("ChannelModel", StringValue ("ns3::MmWaveChannelRaytracing"));
    }
    // ****** The channel matrix model is already obsoleted by the 3GPP Channel Model, please use the later one ******
    // channel matrix model removed from this version
    else if ( channelBeamForming ) {

        // this is real-time, with no traffic travelling over the network
        std::cout << "MmWaveBeamforming is now active\n";
/*
        // Time (ms) between periodic updating of channel matrix/beamforming vectors
        Config::SetDefault ("ns3::MmWaveBeamforming::LongTermUpdatePeriod", TimeValue (MilliSeconds (beam_longTermUpdatePeriod)));
        // Enable small scale fading
        Config::SetDefault ("ns3::MmWaveBeamforming::SmallScaleFading", BooleanValue (smallScale));
        // Set a fixed speed (even if constant position) so doppler > 0 for testing
        Config::SetDefault ("ns3::MmWaveBeamforming::FixSpeed", BooleanValue (beam_fixSpeed));
        // UE speed (m/s) for fixed speed test
        Config::SetDefault ("ns3::MmWaveBeamforming::UeSpeed", DoubleValue (speed));
*/
        // use the MmWaveBeamforming channel model
        mmwaveHelper->SetAttribute ("ChannelModel", StringValue ("ns3::MmWaveBeamforming"));
    }
    else {

        std::cout << "Default - MmWaveBeamforming - is now active\n";

    }


    // #################  BACKGROUND TRAFFIC ###################
    // add some background trafic - NOT COMPLETE AS PER VERSION 1.0.x
    if (backgroundTraffic)
    {
        numberOfStreamingClients=numberOfNodes+1;
    }
    else {
      numberOfStreamingClients=numberOfNodes;
    }


    // #################  BUILDING SETUP ###################
    // taken from "mmwave-simple-building-obstacle.cc" - create a building
    if(enableBuildings) {

        std::cout << "Building creation is now active\n";

        // you can't pass an enum from commandline, so I have to pass a string and
        // based on string, build the enum required

        // set the building type
        Building::BuildingType_t current_building_type;
        // if building is Residential
        if( building_type == "Residential" ) {

            current_building_type = Building::Residential;
        }
        // if building is Office
        else if ( building_type == "Office" ){

            current_building_type = Building::Office;
        }
        // if building is Commercial
        else {

            current_building_type = Building::Commercial;
        }

        // set the wall type
        Building::ExtWallsType_t current_extWalls_type;
        // if wall is ConcreteWithWindows
        if( mmwavePathloss == "ConcreteWithWindows" ) {

            current_extWalls_type = Building::ConcreteWithWindows;
        }
        // if wall is ConcreteWithoutWindows
        else if (mmwavePathloss == "ConcreteWithoutWindows" ){

            current_extWalls_type = Building::ConcreteWithoutWindows;
        }
        // if wall is StoneBlocks
        else {

            current_extWalls_type = Building::StoneBlocks;
        }

        // create a building
        Ptr < Building > building;
        building = Create<Building> ();
        // The boundaries of this Building as a value of type ns3::Box
        // ns3::Box ( xMin, xMax, yMin, yMax, zMin, zMax)
        building->SetBoundaries (Box (xMin, xMax,
                                yMin, yMax,
                                zMin, zMax));
        // The type of building - default Residential - options Residential, Office, Commercial
        building->SetBuildingType (current_building_type);
        // Type of material of which the external walls are made - default ConcreteWithWindows
        // options ConcreteWithWindows, ConcreteWithoutWindows, StoneBlocks
        building->SetExtWallsType (current_extWalls_type);
        // The number of floors of this building - default uint32_t 1
        building->SetNFloors (num_Floors);
        // The number of rooms in the X axis - default uint32_t 1
        building->SetNRoomsX (num_RoomsX);
        // The number of rooms in the Y axis - default uint32_t 1
        building->SetNRoomsY (num_RoomsY);

    }


    // #################  CONFIG SETUP ###################

    std::cout << "Update config settings\n";
    ConfigStore inputConfig;
    inputConfig.ConfigureDefaults();
    // parse again so you can override default values from the command line
    cmd.Parse(argc, argv);

    // *****************  CREATE PROTOCOL STACKS **********************
    Ipv4StaticRoutingHelper ipv4RoutingHelper;
    Ipv4ListRoutingHelper list1;
    list1.Add (ipv4RoutingHelper, 10);

    // *****************  LETS CREATE THE INTERNET SIDE OF THE EPC **********************
    // create our containers - code found online
    std::cout << "\n";
    std::cout << "**** SERVER NODE ****\n";
    std::cout << "Create Server nodes\n";
    NodeContainer remoteNodes;
    remoteNodes.Create (2);
    Ptr<Node> leftTapLeg = remoteNodes.Get (0);
    Ptr<Node> remoteHost = remoteNodes.Get (1);
    Names::Add ("Remote.1", remoteNodes.Get (0));
    Names::Add ("Remote.2", remoteNodes.Get (1));

    // Create the Internet
    std::cout << "Create CSMA & Internet Helpers\n";
    CsmaHelper csmaRemote;
    csmaRemote.SetChannelAttribute ("DataRate", DataRateValue (DataRate (device_DataRate.c_str())));

    NetDeviceContainer internetDevicesRemote = csmaRemote.Install (remoteNodes);
    // lets add some more names
    Names::Add ("Remote_Left_CSMA", internetDevicesRemote.Get (0));
    Names::Add ("Remote_Right_CSMA", internetDevicesRemote.Get (1));
    InternetStackHelper internetRemote;

    // this is need for external ping to work
    //internetRemote.SetRoutingHelper (list1);
    internetRemote.Install (remoteNodes);

    // this is not needed for external ping to work
    //internetRemote.Reset(); //clear the stack

    Ipv4AddressHelper ipv4Remote;
    ipv4Remote.SetBase ("11.0.0.0", "255.255.255.0");
    // burn the IP address, we use
    ipv4Remote.NewAddress();
    ipv4Remote.NewAddress();
    Ipv4InterfaceContainer interfacesLeft = ipv4Remote.Assign (internetDevicesRemote);
    // added
    ipv4Remote.NewNetwork ();

    // get the translation values between the ueNodes and TAP
    Ipv4Address addr8 = interfacesLeft.GetAddress (0);
    std::cout << "Remote Left IP Address = " << addr8 << ".\n";
    Ipv4Address addr7 = interfacesLeft.GetAddress (1);
    std::cout << "Remote Right IP Address = " << addr7 << ".\n";

    //TapBridgeHelper tapBridgeRemote;
    TapBridgeHelper tapBridgeRemote(interfacesLeft.GetAddress (1));
    // UseBridge is used for extending a logical bridge into ns3
    tapBridgeRemote.SetAttribute ("Mode", StringValue ("UseBridge"));
    tapBridgeRemote.SetAttribute ("DeviceName", StringValue ("tapNas"));
    if (!trafficGeneration)
    {
      tapBridgeRemote.Install (leftTapLeg, internetDevicesRemote.Get (0));
    }
    
    // adding global routing
    //Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

    // get PGW node
    Ptr<Node> pgw = epcHelper->GetPgwNode ();
    // lets add a name
    Names::Add ("PGW", pgw);

    std::cout << "\n";
    std::cout << "Create P2P Helper\n";
    PointToPointHelper p2ph;
    p2ph.SetDeviceAttribute ("DataRate", DataRateValue (DataRate (device_DataRate.c_str())));
    p2ph.SetDeviceAttribute ("Mtu", UintegerValue (2500));
    p2ph.SetChannelAttribute ("Delay", TimeValue (Seconds (channel_Delay)));
    Ipv4AddressHelper ipv4h;
    ipv4h.SetBase ("1.0.0.0", "255.255.255.0");
    NetDeviceContainer internetDevices = p2ph.Install (pgw, remoteHost);

    Ipv4InterfaceContainer internetIpIfaces = ipv4h.Assign (internetDevices);
    ipv4h.NewNetwork ();

    // lets add some more names
    Names::Add ("Remote_P2P", internetDevices.Get (1));
    Names::Add ("Pgw_P2P", internetDevices.Get (0));

    Ipv4Address addr9 = internetIpIfaces.GetAddress (1);
    std::cout << "remote Host IP = " << addr9 << ".\n";
    Ipv4Address addr10 = internetIpIfaces.GetAddress (0);
    std::cout << "pgw IP = " << addr10 << ".\n";

    // adding global routing
    //Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

    // *****************  LETS CREATE THE UE AND ENB NODES **********************
    std::cout << "\n";
    std::cout << "Create UE and ENB nodes\n";
    NodeContainer ueNodes;
    NodeContainer enbNodes;
    enbNodes.Create(numberOfeNodeBs);
    Names::Add ("ENB", enbNodes.Get (0));
    ueNodes.Create(numberOfStreamingClients);

    // add some names to the ueNodes
    for (uint16_t i = 0; i < numberOfStreamingClients; i++)
    {
        std::stringstream ueNodesName;
        ueNodesName << "UE" << (i+1) << ".1";
        Names::Add (ueNodesName.str(), ueNodes.Get (i));
    }

    std::cout << "Install Mobility Model\n";
    Ptr<ListPositionAllocator> enbPositionAlloc = CreateObject<ListPositionAllocator> ();
    enbPositionAlloc->Add (Vector (0.0, 0.0, hBS));
    MobilityHelper mobility;
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.SetPositionAllocator(enbPositionAlloc);
    mobility.Install (enbNodes);

    Ptr<Node> enb1 = enbNodes.Get (0);
    Ptr<MobilityModel> mobModelenb1 = enb1->GetObject<MobilityModel> ();
    Vector positionenb1 = mobModelenb1->GetPosition();
    std::cout<< "The position of ENB1 is "<<positionenb1<<std::endl;

    // taken from "mmwave-simple-building-obstacle.cc"
    if(enableBuildings) {

        // add enbNodes to BuildingHelper
	    BuildingsHelper::Install (enbNodes);

    }

    // bound the box based on maxDistace from Base Station
    int boxBound = maxDistance;
    // implement mobilty only if we use a mobile fading model and we want to use mobility
    if (mobileFading && useMobility)
    {

        if (vehicleFading){
            std::cout << "Vehicular mobility ..... "  << ".\n";
            mobility.SetMobilityModel ("ns3::GaussMarkovMobilityModel",
           "Bounds", BoxValue (Box (-1*boxBound, boxBound, -1*boxBound, boxBound, 0, 100)),
          "TimeStep", TimeValue (Seconds (0.5)),
          "Alpha", DoubleValue (0.85),
          // I won't remove pitch for vehicle mobility, as the road may not be straight
          "MeanVelocity", StringValue ("ns3::ConstantRandomVariable[Constant=8.33]"),
          "MeanDirection", StringValue ("ns3::UniformRandomVariable[Min=0|Max=6.283185307]"),
          "MeanPitch", StringValue ("ns3::UniformRandomVariable[Min=0.005|Max=0.005]"),
          "NormalVelocity", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=1.0|Bound=12.0]"),
          "NormalDirection", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.2|Bound=3.14]"),
          "NormalPitch", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.0002|Bound=0.0004]"));

        } else {
            std::cout << "Pedestrian mobility ..... "  << ".\n";
            mobility.SetMobilityModel ("ns3::GaussMarkovMobilityModel",
          "Bounds", BoxValue (Box (-1*boxBound, boxBound, -1*boxBound, boxBound, 0, 100)),
          "TimeStep", TimeValue (Seconds (0.5)), // this can be used to vary the printed output
          "Alpha", DoubleValue (0.85),
          "MeanVelocity", StringValue ("ns3::ConstantRandomVariable[Constant=1]"),
          "MeanDirection", StringValue ("ns3::UniformRandomVariable[Min=0|Max=6.283185307]"),
          // remove pitch - not needed for pedestrian?
          // "MeanPitch", StringValue ("ns3::UniformRandomVariable[Min=0.005|Max=0.005]"),
          "MeanPitch", StringValue ("ns3::UniformRandomVariable[Min=0.00|Max=0.00]"),
          "NormalVelocity", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=1.0|Bound=12.0]"),
          "NormalDirection", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.2|Bound=3.14]"),
          // remove pitch - not needed for pedestrian?
          // "NormalPitch", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.0002|Bound=0.0004]"
          "NormalPitch", StringValue ("ns3::NormalRandomVariable[Mean=0.0|Variance=0.0000|Bound=0.0000]"));

        }

    } else {

        mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel"); // probably no need for this
        Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
        for (uint16_t i = 0; i < numberOfNodes; i++)
            {
                // changed y position of the device from 0 to i*2 - improve LoS
                positionAlloc->Add (Vector(distance[i], i*2, hUT)); // different to D-LiTE
                std::cout << "Mobility allocated for UE-" << i+1 << " at a distance of " << distance[i] << ".\n";
            }

        mobility.SetPositionAllocator(positionAlloc);
    }
    mobility.Install(ueNodes);

    // update mobilty only if we use a mobile fading model and we want to use mobility
    if (mobileFading && useMobility)
    {
        for (uint32_t i = 0; i < ueNodes.GetN (); i++)
        {
            Ptr<MobilityModel> ueMmPtr = (ueNodes.Get(i))->GetObject<MobilityModel>();

            // lets randomise the values between -maxDistance and +maxDistance
            double xPos = (rand()%((int(maxDistance)*2)+1)-int(maxDistance)) * cos(0.785);
            double yPos = (rand()%((int(maxDistance)*2)+1)-int(maxDistance)) * sin(0.785);

            // original code
            //double xPos = (40 + i * 70) * cos(0.785);
            //double yPos = (40 + i * 70) * sin(0.785);
            ueMmPtr->SetPosition(Vector(xPos, yPos, hUT));
        }
    }
    /*
    for (uint16_t i = 0; i <ueNodes.GetN (); i++)
    {
        Ptr<Node> UE = ueNodes.Get (i);
        Ptr<MobilityModel> mobModelUE1 = UE->GetObject<MobilityModel> ();
        Vector positionUE = mobModelUE1->GetPosition();
        std::cout<< "The position of UE"<<i+1<<" is "<<positionUE<<std::endl;
    }
*/
    // taken from "mmwave-simple-building-obstacle.cc"
    if(enableBuildings) {

        // add enbNodes to BuildingHelper
	    BuildingsHelper::Install (ueNodes);

    }

    // existing - Install LTE Devices to the nodes
    std::cout << "\nCreate UE and ENB Containers\n";
    NetDeviceContainer enbMmWaveDevs = mmwaveHelper->InstallEnbDevice (enbNodes);
    // lets add some more names
    Names::Add ("ENB1", enbMmWaveDevs.Get (0));

/*
    // ****** to change ENB bandwidth ******
    Ptr<MmWaveEnbNetDevice> enbMmWaveDevice = enbmmWaveDevs.Get (0)->GetObject<MmWaveEnbNetDevice> ();
    // this works to change enb bandwidth,but does not change UE bandwidth
    enbMmWaveDevice->SetBandwidth(36);
*/

    NetDeviceContainer ueMmWaveDevs = mmwaveHelper->InstallUeDevice (ueNodes);

    // add some names to the ueMmWaveDevs
    for (uint16_t i = 0; i < numberOfStreamingClients; i++)
    {
        // string conversion using IOStreams
        std::stringstream ueMmwaveDevsName;
        ueMmwaveDevsName << "LTE_UE" << (i+1);
        // lets add some more names
        Names::Add (ueMmwaveDevsName.str(), ueMmWaveDevs.Get (i));
    }

    // Install the IP stack on the UEs
    std::cout << "Install Internet Stack on UEs\n";
    InternetStackHelper internet;
    //internet.Install (ueNodes);

    //install protocol stack to the UEs
    for (uint32_t i = 0; i < ueNodes.GetN (); i++)
    {
       internet.SetRoutingHelper (list1);
       internet.Install (ueNodes.Get(i));
     }
    internet.Reset();

    Ipv4InterfaceContainer ueIpIface;
    ueIpIface = epcHelper->AssignUeIpv4Address (NetDeviceContainer (ueMmWaveDevs));

    // Assign IP address to UEs, and install applications
    for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
    {
        Ptr<Node> ueNode = ueNodes.Get (u);
        // Set the default gateway for the UE
        Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ueNode->GetObject<Ipv4> ());
        ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);
    }

    // attach the Ue to the closets ENB
    mmwaveHelper->AttachToClosestEnb (ueMmWaveDevs, enbMmWaveDevs);

    // *****************  LETS CREATE SOME UE'S ************************
    // Let's create the first node
    std::cout << "\n";
    std::cout << "Create IP Address Helper\n";
    ipv4h.SetBase ("18.2.0.0", "255.255.0.0");
    // burn the IP address, we use
    //ipv4h.NewAddress();
    // burn the IP address, we use
    std::cout << "View Network Metrics\n";
    Ipv4Address gateway = epcHelper->GetUeDefaultGatewayAddress ();
    std::cout << "Left UE gateway = " << gateway << ".\n";

    // create the UE nodes
    NodeContainer UE;
    UE.Create (numberOfStreamingClients);

    // creat the csma
    CsmaHelper csma;
    csma.SetChannelAttribute ("DataRate", DataRateValue (DataRate (device_DataRate.c_str())));

    // create the tap bridge
    TapBridgeHelper tapBridge;
    tapBridge.SetAttribute ("Mode", StringValue ("UseBridge"));

    Ipv4InterfaceContainer interfacesUE;

    // add some UEs - based on  the number of devices - this will be one less when background traffic is true
    for (uint16_t i = 0; i < numberOfNodes; i++)
    {
        // string conversion using IOStreams
        std::stringstream ueFirstName;
        std::stringstream ueLeftName;
        std::stringstream ueRightName;
        std::stringstream ueTapName;

        // create the strings
        ueFirstName << "UE" << (i+1) << ".2";
        ueLeftName << "UE" << (i+1) << "_Left_CSMA";
        ueRightName << "UE" << (i+1) << "_Right_CSMA";
        ueTapName << "tap-UE" << (i+1);

        // create this UE
        std::cout << "**** UE-"<<(i+1)<<" ****\n";
        NetDeviceContainer internetDevice = csma.Install (NodeContainer (ueNodes.Get (i), UE.Get (i)));
        internet.Install (UE.Get (i));
        // lets add some more names
        Names::Add (ueFirstName.str(), UE.Get (i));
        Names::Add (ueLeftName.str(), internetDevice.Get (0));
        Names::Add (ueRightName.str(), internetDevice.Get (1));
        interfacesUE = ipv4h.Assign (internetDevice);
        std::cout << "Create Tap Bridge\n";
        
        if (!trafficGeneration)
        {
          tapBridge.SetAttribute ("DeviceName", StringValue (ueTapName.str()));
          tapBridge.Install (UE.Get (i), internetDevice.Get (1));
        }
        
     }

    // this will be one less when background traffic is true
    for (uint32_t u = 0; u < numberOfNodes; ++u)
    {
        std::cout << "\n";
        // get the translation values between the ueNodes and TAP
        Ptr<Node> ueNode = ueNodes.Get (u);
        Ptr<Ipv4> ipv4 = ueNode->GetObject<Ipv4> ();
        Ipv4Address addr1 = ipv4->GetAddress (1, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " LTE IP Address = " << addr1 << ".\n";
        Ipv4Address addr2 = ipv4->GetAddress (2, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " LTE-CSMA IP Address = " << addr2 << ".\n";
        Ipv4Address TapAddr = UE.Get (u)->GetObject<Ipv4> ()->GetAddress (1, 0).GetLocal ();
        std::cout << "UE-" << u+1 << " CSMA-TAP IP Address = " << TapAddr << ".\n";

        Ptr<MmWaveUeNetDevice> ueMmWaveDevice = ueMmWaveDevs.Get (u)->GetObject<MmWaveUeNetDevice> ();
        uint64_t ueImsi = ueMmWaveDevice->GetImsi ();
        std::cout << "UE-" << u+1 << " IMSI = " << ueImsi << ".\n";
    }

    // add some background trafic
    if (backgroundTraffic)
    {
        // increase the number of UE
        // numberOfNodes=numberOfNodes+1;
    }

//    // ading global routing
//    Ipv4GlobalRoutingHelper::PopulateRoutingTables ();

    // *****************  LETS CREATE SOME ROUTING ***********************
    std::cout << "\n";
    std::cout << "Create Routing\n";
    // define some static routes for the REMOTE NODE
    Ptr<Ipv4StaticRouting> remoteHostStaticRouting = ipv4RoutingHelper.GetStaticRouting (remoteHost->GetObject<Ipv4> ());
    // 9.0.0.x (NAS) and 12.0.0.x (Clients) are subnets to LTE and the Clients
    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("11.0.0.0"), Ipv4Mask ("255.255.255.0"), 1);
    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("9.0.0.0"), Ipv4Mask ("255.255.255.0"), Ipv4Address ("11.0.0.1"), 1);

    remoteHostStaticRouting->AddNetworkRouteTo (Ipv4Address ("18.2.0.0"), Ipv4Mask ("255.0.0.0"), 2);


    // define some static routes for the PGW
    Ptr<Ipv4StaticRouting> remoteHostStaticRoutingPgw = ipv4RoutingHelper.GetStaticRouting (pgw->GetObject<Ipv4> ());
    remoteHostStaticRoutingPgw->AddNetworkRouteTo (Ipv4Address ("9.0.0.0"), Ipv4Mask ("255.255.255.0"), 2);
    remoteHostStaticRoutingPgw->AddNetworkRouteTo (Ipv4Address ("11.0.0.0"), Ipv4Mask ("255.255.255.0"), 2);


   // only add routing for internal nodes when we want internal traffic
   if (trafficGeneration)
   {
		  // this will be one less when background traffic is true
       for (uint32_t u = 0; u < numberOfNodes; ++u)
       {
          // Set the default gateway for the UE
	        Ptr<Node> ueNode = ueNodes.Get (u);
	        Ptr<Ipv4StaticRouting> ueStaticRouting = ipv4RoutingHelper.GetStaticRouting (ueNode->GetObject<Ipv4> ());
	        ueStaticRouting->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);
          ueStaticRouting->AddNetworkRouteTo (Ipv4Address ("18.2.0.0"), Ipv4Mask ("255.0.0.0"), 3);

          // only used when we don't use TapBridge
          Ptr<Node> ueNodeRemote = UE.Get (u);
          Ptr<Ipv4StaticRouting> ueStaticRoutingRemote = ipv4RoutingHelper.GetStaticRouting (ueNodeRemote->GetObject<Ipv4> ());
          // ueStaticRoutingRemote->SetDefaultRoute (epcHelper->GetUeDefaultGatewayAddress (), 1);

          std::ostringstream subnet;
          subnet<<"18.2.0."<<((u*2)+1);
          std::cout << "IP of gateway " << subnet.str ().c_str () << " for client " << u+1 << "\n";
          ueStaticRoutingRemote->AddNetworkRouteTo (Ipv4Address ("1.0.0.0"), Ipv4Mask ("255.255.255.0"), Ipv4Address (subnet.str ().c_str ()), 1);

       }
   }

	// nice way to dynamically adapt ip addresses
//	Ipv4AddressHelper ipv4h;
//    std::ostringstream subnet;
//    subnet<<i<<".1.0.0";
//	ipv4h.SetBase (subnet.str ().c_str (), "255.255.0.0");
//	internetIpIfaces = ipv4h.Assign (internetDevices);


    // *****************  PRODUCE SOME TRACES AND START THE SIMULATION  *************************

    /*
    // flowMonitoring for PacketFlows
    Ptr<FlowMonitor> flowmon;
    if (flowMonitor)
    {
      FlowMonitorHelper flowmonHelper;
      flowmon = flowmonHelper.InstallAll ();
    }

    // Activate a data radio bearer
    enum EpsBearer::Qci q = EpsBearer::GBR_CONV_VOICE;
    EpsBearer bearer (q);
    mmwaveHelper->ActivateDataRadioBearer (ueMmWaveDevs, bearer);
*/
    // *****************  GET SOME VALUES FROM THE SIMULATION  *************************
    for (uint32_t u = 0; u < numberOfStreamingClients; ++u)
    {
        // get bandwidth values for each UE, from within the running simulation
        // Simulator::Schedule ( Seconds(0.5+u), &checkBandwidth , ueMmWaveDevs.Get (u), enbMmWaveDevs.Get (0));

    }

    // reset the contents of our stop file
    Simulator::Schedule ( Seconds(0), &writeSimulationValue );

    // check the file contents and stop the simulator if equal to 1
    for (uint32_t u = 1; u <= simTime; ++u)
    {
        // stop the simulator
        Simulator::Schedule ( Seconds(u), &checkSimulation, u );
    }


    // ***************** ANIMATION *****************
    // Create the animation object and configure for specified output
    // AnimationInterface anim (animFile);
    // anim.EnablePacketMetadata (); // Optional
    // anim.EnableIpv4L3ProtocolCounters (Seconds (0), Seconds (10)); // Optional


    // from lena-x2-handover-measures.cc
    // connect custom trace sinks for RRC connection establishment and handover notification
    Config::Connect ("/NodeList/*/DeviceList/*/LteUeRrc/ConnectionEstablished",
                   MakeCallback (&NotifyConnectionEstablishedEnb));


    // only change direction if we want to use mobility
    if( usePrintMobility){
        Config::Connect ("/NodeList/*/$ns3::MobilityModel/CourseChange",
                   MakeCallback (&CourseChange));
    }


    // taken from "mmwave-3gpp-channel-example.cc" - enable the building propagation
    if(enableBuildings) {

        //
	    BuildingsHelper::MakeMobilityModelConsistent ();

    }

    // *****************  START THE SIMULATION  *************************
    // if we want to generate traffic
    if (trafficGeneration) {

        std::cout << "\n";
        // if we want to generate wireshark pcap files
        if (pcapGenerate) {
            std::cout << "Enable Traces\n";
            // CREATE WIRESHARK TRACES
            p2ph.EnablePcapAll("MMWAVE_P2P_Link");
            csma.EnablePcapAll ("MMWAVE_CSMA_Devices");
        }
        // from D-LiTE
        // these enable some key performance indicators (KPIs)
        // files are created and these are view in the root ns3 folder
        // see src/lte/doc/source/lte-user.rst - line 196
        // lteHelper->EnableMacTraces();
        // lteHelper->EnableRlcTraces();
        // lteHelper->EnablePdcpTraces();

        // if we want TCP traffic, if not then UDP traffic
        if(tcp)
	      {

	        // Install and start applications on UEs and remote host
	        uint16_t sinkPort = 20000;

            // from ue node 18.0.0.2
	        // Address sinkAddress (InetSocketAddress (ueIpIface.GetAddress (0), sinkPort));
	        // PacketSinkHelper packetSinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), sinkPort));
	        // ApplicationContainer sinkApps = packetSinkHelper.Install (ueNodes.Get (0));

            // from ue node 18.1.0.3
            Address sinkAddress (InetSocketAddress (interfacesUE.GetAddress (1), sinkPort));
	        PacketSinkHelper packetSinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), sinkPort));
            // the more UE you add, you need to increase this to numberOfStreamingClients less 1
	        ApplicationContainer sinkApps = packetSinkHelper.Install (UE.Get (numberOfStreamingClients-1));


	        sinkApps.Start (Seconds (0.0));
	        sinkApps.Stop (Seconds (5));

	        Ptr<Socket> ns3TcpSocket = Socket::CreateSocket (remoteNodes.Get (1), TcpSocketFactory::GetTypeId ());
	        Ptr<MyApp> app = CreateObject<MyApp> ();
	        app->Setup (ns3TcpSocket, sinkAddress, 1400, 5000000, DataRate ("1000Mb/s"));

	        remoteNodes.Get (1)->AddApplication (app);
	        AsciiTraceHelper asciiTraceHelper;
	        Ptr<OutputStreamWrapper> stream1 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-window-newreno.txt");
	        ns3TcpSocket->TraceConnectWithoutContext ("CongestionWindow", MakeBoundCallback (&CwndChange, stream1));

	        Ptr<OutputStreamWrapper> stream4 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-rtt-newreno.txt");
	        ns3TcpSocket->TraceConnectWithoutContext ("RTT", MakeBoundCallback (&RttChange, stream4));

	        Ptr<OutputStreamWrapper> stream2 = asciiTraceHelper.CreateFileStream ("mmWave-tcp-data-newreno.txt");
	        sinkApps.Get(0)->TraceConnectWithoutContext("Rx",MakeBoundCallback (&Rx, stream2));

	        app->SetStartTime (Seconds (0.1));
	        app->SetStopTime (Seconds (5));

	    }
	    else
	    {

            Ipv4Address remoteHostAddr = internetIpIfaces.GetAddress (1);
        // Install and start applications on UEs and remote host
        uint16_t dlPort = 1234;
        uint16_t ulPort = 2000;
        uint16_t otherPort = 3000;
        ApplicationContainer clientApps;
        ApplicationContainer serverApps;
        for (uint32_t u = 0; u < ueNodes.GetN (); ++u)
        {
          ++ulPort;
          ++otherPort;
          PacketSinkHelper dlPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), dlPort));
          PacketSinkHelper ulPacketSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), ulPort));
          PacketSinkHelper packetSinkHelper ("ns3::UdpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), otherPort));
          serverApps.Add (dlPacketSinkHelper.Install (UE.Get(u)));
          serverApps.Add (ulPacketSinkHelper.Install (remoteHost));
          serverApps.Add (packetSinkHelper.Install (UE.Get(u)));

          UdpClientHelper dlClient (interfacesUE.GetAddress (1), dlPort);
          dlClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          dlClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

          UdpClientHelper ulClient (remoteHostAddr, ulPort);
          ulClient.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          ulClient.SetAttribute ("MaxPackets", UintegerValue(1000000));

          UdpClientHelper client (interfacesUE.GetAddress (1), otherPort);
          client.SetAttribute ("Interval", TimeValue (MilliSeconds(interPacketInterval)));
          client.SetAttribute ("MaxPackets", UintegerValue(1000000));

          clientApps.Add (dlClient.Install (remoteHost));
          clientApps.Add (ulClient.Install (UE.Get(u)));
          if (u+1 < UE.GetN ())
            {
              clientApps.Add (client.Install (UE.Get(u+1)));
            }
          else
            {
              clientApps.Add (client.Install (UE.Get(0)));
            }
          }
         serverApps.Start (Seconds (0.01));
         clientApps.Start (Seconds (0.01));
        }
      }

    std::cout << "Start Simulator\n";
    Simulator::Stop(Seconds(simTime));

    // start real-time check
    startTime = std::chrono::system_clock::now();
    Simulator::Run();
/*
    if (flowMonitor)
    {

      // 2nd and 3rd paameters are used to turn on/off histograms and per-probe detailed stats
      flowmon->SerializeToXmlFile ("./5GflowMonitor.xml", true, false);
    }
*/

    std::cout << "\n";
    std::cout << "Done\n";

    Simulator::Destroy();
    return 0;

}
